#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Manager class for a sequential optical model

.. codeauthor: Michael J. Hayford
"""

import itertools
import logging

from . import surface
from . import gap
from . import medium as m
from . import raytrace as rt
from . import trace as trace
from . import transform as trns
from rayoptics.optical.model_constants import Intfc, Gap, Indx, Tfrm, Zdir
from opticalglass import glassfactory as gfact
from opticalglass import glasserror as ge
import numpy as np
from math import copysign, sqrt
from rayoptics.util.misc_math import isanumber


class SequentialModel:
    """ Manager class for a sequential optical model

    A sequential optical model is a sequence of surfaces and gaps.

    The sequential model has this structure
    ::

        IfcObj  Ifc1  Ifc2  Ifc3 ... Ifci-1   IfcImg
             \  /  \  /  \  /             \   /
             GObj   G1    G2              Gi-1

    where

        - Ifc is a :class:`~rayoptics.optical.surface.Interface` instance
        - G   is a :class:`~rayoptics.optical.gap.Gap` instance

    There are N interfaces and N-1 gaps. The initial configuration has an
    object and image Surface and an object gap.

    The Interface API supports implementation of an optical action, such as
    refraction, reflection, scatter, diffraction, etc. The Interface may be
    realized as a physical profile separating the adjacent gaps or an idealized
    object, such as a thin lens or 2 point HOE.
    The Gap class maintains a simple separation (z translation) and the medium
    filling the gap. More complex coordinate transformations are handled
    through the Interface API.

    Attributes:
        opt_model: parent optical model
        ifcs: list of :class:`~rayoptics.optical.surface.Interface`
        gaps: list of :class:`~rayoptics.optical.gap.Gap`
        lcl_tfrms: forward transform, interface to interface
        rndx: a list with refractive indices for all **wvls**
        z_dir: -1 if gap follows an odd number of reflections, otherwise +1
        gbl_tfrms: global coordinates of each interface wrt the 1st interface
        stop_surface (int): index of stop interface
        cur_surface (int): insertion index for next interface
    """

    def __init__(self, opt_model):
        self.opt_model = opt_model
        self.ifcs = []
        self.gaps = []
        self.gbl_tfrms = []
        self.lcl_tfrms = []
        self.z_dir = []
        self.stop_surface = 1
        self.cur_surface = 0
        self.wvlns = []
        self.rndx = []
        self._initialize_arrays()

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['gbl_tfrms']
        del attrs['lcl_tfrms']
        del attrs['z_dir']
        del attrs['wvlns']
        del attrs['rndx']
        return attrs

    def _initialize_arrays(self):
        """ initialize object and image interfaces and intervening gap """
        # add object interface
        self.ifcs.append(surface.Surface('Obj'))

        tfrm = np.identity(3), np.array([0., 0., 0.])
        self.gbl_tfrms.append(tfrm)
        self.lcl_tfrms.append(tfrm)

        # add object gap
        self.gaps.append(gap.Gap())
        self.z_dir.append(1)
        self.rndx.append([1.0])

        # add image interface
        self.ifcs.append(surface.Surface('Img'))
        self.gbl_tfrms.append(tfrm)
        self.lcl_tfrms.append(tfrm)
        self.z_dir.append(1)
        self.rndx.append([1.0])

    def reset(self):
        self.__init__()

    def get_num_surfaces(self):
        return len(self.ifcs)

    def path(self, wl=None, start=None, stop=None, step=1):
        """ returns an iterable path tuple for a range in the sequential model

        Args:
            wl: wavelength in nm for path, defaults to central wavelength
            start: start of range
            stop: first value beyond the end of the range
            step: increment or stride of range

        Returns:
            (**ifcs, gaps, lcl_tfrms, rndx, z_dir**)
        """
        if wl is None:
            wl = self.central_wavelength()

        if step < 0:
            gap_start = start - 1
        else:
            gap_start = start

        wl_idx = self.index_for_wavelength(wl)
        rndx = [n[wl_idx] for n in self.rndx[start:stop:step]]
        path = itertools.zip_longest(self.ifcs[start:stop:step],
                                     self.gaps[gap_start:stop:step],
                                     self.lcl_tfrms[start:stop:step],
                                     rndx,
                                     self.z_dir[start:stop:step])
        return path

    def calc_ref_indices_for_spectrum(self, wvls):
        """ returns a list with refractive indices for all **wvls**

        Args:
            wvls: list of wavelengths in nm
        """
        indices = []
        for g in self.gaps:
            ri = []
            mat = g.medium
            for w in wvls:
                rndx = mat.rindex(w)
                ri.append(rndx)
            indices.append(ri)
        indices.append(indices[-1])

        return indices

    def central_wavelength(self):
        """ returns the central wavelength in nm of the model's ``WvlSpec`` """
        return self.opt_model.optical_spec.spectral_region.central_wvl

    def index_for_wavelength(self, wvl):
        """ returns index into rndx array for wavelength `wvl` in nm """
        if len(self.wvlns) == 0:
            rgn = self.opt_model.optical_spec.spectral_region
            self.wvlns = rgn.wavelengths

        return self.wvlns.index(wvl)

    def central_rndx(self, i):
        """ returns the central refractive index of the model's ``WvlSpec`` """
        central_wvl = self.opt_model.optical_spec.spectral_region.reference_wvl
        return self.rndx[i][central_wvl]

    def get_surface_and_gap(self, srf=None):
        if not srf:
            srf = self.cur_surface
        s = self.ifcs[srf]
        if srf == len(self.gaps):
            g = None
        else:
            g = self.gaps[srf]
        return s, g

    def set_cur_surface(self, s):
        self.cur_surface = s

    def set_stop(self):
        """ sets the stop surface to the current surface """
        self.stop_surface = self.cur_surface
        return self.stop_surface

    def __iadd__(self, node):
        if isinstance(node, gap.Gap):
            self.gaps.append(node)
        else:
            self.ifcs.insert(len(self.ifcs)-1, node)
        return self

    def insert(self, ifc, gap):
        """ insert surf and gap at the cur_gap edge of the sequential model
            graph """
        if len(self.ifcs) > 2:
            if self.stop_surface is not None:
                if self.stop_surface > self.cur_surface:
                    self.stop_surface += 1
        self.cur_surface += 1
        surf = self.cur_surface
        self.ifcs.insert(surf, ifc)
        if gap is not None:
            self.gaps.insert(surf, gap)
        else:
            gap = self.gaps[surf]

        tfrm = np.identity(3), np.array([0., 0., 0.])
        self.gbl_tfrms.insert(surf, tfrm)
        self.lcl_tfrms.insert(surf, tfrm)

        self.z_dir.insert(surf, self.z_dir[surf-1])

        wvls = self.opt_model.optical_spec.spectral_region.wavelengths
        rindex = [gap.medium.rindex(w) for w in wvls]
        self.rndx.insert(surf, rindex)

        if ifc.interact_mode == 'reflect':
            self.update_reflections(start=surf)

    def remove(self, *args):
        """ remove surf and gap at cur_surface or an input index argument """
        if len(args) == 0:
            idx = self.cur_surface
        else:
            idx = args[0]

        # don't allow object or image interfaces to be removed
        if idx == 0 or idx == -1 or idx == len(self.ifcs):
            raise IndexError

        if self.ifcs[idx].interact_mode == 'reflect':
            self.update_reflections(start=idx)

        del self.ifcs[idx]
        del self.gaps[idx]
        del self.gbl_tfrms[idx]
        del self.lcl_tfrms[idx]
        del self.z_dir[idx]
        del self.rndx[idx]

    def add_surface(self, surf_data, **kwargs):
        """ add a surface where surf is a list that contains:
            [curvature, thickness, refractive_index, v-number] """
        radius_mode = self.opt_model.radius_mode
        mat = None
        if len(surf_data) > 2:
            if not isanumber(surf_data[2]):
                if surf_data[2].upper() == 'REFL':
                    mat = self.gaps[self.cur_surface-1].medium
        s, g, rn, tfrm = create_surface_and_gap(surf_data, prev_medium=mat,
                                                radius_mode=radius_mode,
                                                **kwargs)
        self.insert(s, g)

    def sync_to_restore(self, opt_model):
        self.opt_model = opt_model
        if hasattr(self, 'optical_spec'):
            opt_model.optical_spec = self.optical_spec
            delattr(self, 'optical_spec')
        for sg in itertools.zip_longest(self.ifcs, self.gaps):
            if hasattr(sg[Intfc], 'sync_to_restore'):
                sg[Intfc].sync_to_restore(self)
            if sg[Gap]:
                if hasattr(sg[Gap], 'sync_to_restore'):
                    sg[Gap].sync_to_restore(self)

    def update_model(self):
        # delta n across each surface interface must be set to some
        #  reasonable default value. use the index at the central wavelength
        osp = self.opt_model.optical_spec
        ref_wl = osp.spectral_region.reference_wvl

        self.wvlns = osp.spectral_region.wavelengths
        self.rndx = self.calc_ref_indices_for_spectrum(self.wvlns)
        n_before = self.rndx[0]

        self.z_dir = []
        z_dir_before = 1

        for i, s in enumerate(self.ifcs):
            z_dir_after = copysign(1, z_dir_before)
            if s.interact_mode == 'reflect':
                z_dir_after = -z_dir_after

            # leave rndx data unsigned, track change of sign using z_dir
            n_after = self.rndx[i]
            if z_dir_after < 0:
                n_after = [-n for n in n_after]
            s.delta_n = n_after[ref_wl] - n_before[ref_wl]
            n_before = n_after

            z_dir_before = z_dir_after
            self.z_dir.append(z_dir_after)
            # call update() on the surface interface
            s.update()

        self.gbl_tfrms = self.compute_global_coords()
        self.lcl_tfrms = self.compute_local_transforms()

        if len(self.ifcs) > 2:
            osp.update_model()

            self.set_clear_apertures()

    def apply_scale_factor(self, scale_factor):
        for i, sg in enumerate(self.path()):
            sg[Intfc].apply_scale_factor(scale_factor)
            if sg[Gap]:
                sg[Gap].apply_scale_factor(scale_factor)

        self.gbl_tfrms = self.compute_global_coords()
        self.lcl_tfrms = self.compute_local_transforms()

    def set_from_specsheet(self, specsheet):
        if self.opt_model.optical_spec.parax_data is None:
            return
        if len(specsheet.imager_inputs) == 2:
            fod = self.opt_model.optical_spec.parax_data.fod
            f_old = fod.efl
            f_new = specsheet.imager.f
            scale_factor = f_new/f_old
            if scale_factor != 1.0:
                self.apply_scale_factor(scale_factor)

            if specsheet.conjugate_type == 'finite':
                self.gaps[0].thi = scale_factor*fod.pp1 - specsheet.imager.s
                self.gaps[-1].thi = specsheet.imager.sp - scale_factor*fod.ppk
            elif specsheet.conjugate_type == 'infinite':
                self.gaps[-1].thi = specsheet.imager.sp - scale_factor*fod.ppk

    def insert_surface_and_gap(self):
        s = surface.Surface()
        g = gap.Gap()
        self.insert(s, g)
        return s, g

    def update_reflections(self, start):
        """ update interfaces and gaps following insertion of a mirror """

        for i, sg in enumerate(self.path(start=start), start=start):
            if i > start:
                sg[Intfc].apply_scale_factor(-1)
                if sg[Gap]:
                    sg[Gap].apply_scale_factor(-1)
            self.z_dir[i] = -sg[Zdir]

    def surface_label_list(self):
        """ list of surface labels or surface number, if no label """
        labels = []
        for i, s in enumerate(self.ifcs):
            if len(s.label) == 0:
                if i == self.stop_surface:
                    labels.append('Stop')
                else:
                    labels.append(str(i))
            else:
                labels.append(s.label)
        return labels

    def list_model(self):
        cvr = 'r' if self.opt_model.radius_mode else 'c'
        print("             {}            t        medium     mode   zdr"
              "      sd".format(cvr))
        labels = self.surface_label_list()
        for i, sg in enumerate(self.path()):
            s = self.list_surface_and_gap(sg[Intfc], gp=sg[Gap])
            s.append(self.z_dir[i])
            print("{0:>4s}: {1:12.6f} {2:#12.6g} {3:>9s} {4:>10s} {6:2n}"
                  "  {5:#10.5g}".format(labels[i], *s))

    def list_model_old(self):
        cvr = 'r' if self.opt_model.radius_mode else 'c'
        print("           {}            t        medium     mode         sd"
              .format(cvr))
        for i, sg in enumerate(self.path()):
            s = self.list_surface_and_gap(sg[Intfc], gp=sg[Gap])
            print("{0:2n}: {1:12.6f} {2:#12.6g} {3:>9s} "
                  "{4.name:>10s} {5:#10.5g}"
                  .format(i, *s))

    def list_gaps(self):
        for i, gp in enumerate(self.gaps):
            print(i, gp)

    def list_surfaces(self):
        for i, s in enumerate(self.ifcs):
            print(i, s)

    def list_surface_and_gap(self, ifc, gp=None):
        cvr = ifc.profile_cv
        if self.opt_model.radius_mode:
            if cvr != 0.0:
                cvr = 1.0/cvr
        sd = ifc.surface_od()
        imode = ifc.interact_mode

        if gp is not None:
            thi = gp.thi
            med = gp.medium.name()
        else:
            thi = 0.
            med = ''
        return [cvr, thi, med, imode, sd]

    def list_decenters(self):
        for i, sg in enumerate(self.path()):
            if sg[Gap]:
                print(i, sg[Gap])
                if sg[Intfc].decenter is not None:
                    print(' ', repr(sg[Intfc].decenter))
            else:
                if sg[Intfc].decenter is not None:
                    print(i, repr(sg[Intfc].decenter))

    def list_elements(self):
        for i, gp in enumerate(self.gaps):
            if gp.medium.label.lower() != 'air':
                print(self.ifcs[i].profile,
                      self.ifcs[i+1].profile,
                      gp)

    def trace_fan(self, fct, fi, xy, num_rays=21):
        """ xy determines whether x (=0) or y (=1) fan """
        osp = self.opt_model.optical_spec
        fld = osp.field_of_view.fields[fi]
        wvl = self.central_wavelength()
        foc = osp.defocus
#        trace.setup_canonical_coords(self, fld, wvl)
        rs_pkg, cr_pkg = trace.setup_pupil_coords(self.opt_model,
                                                  fld, wvl, foc)
        fld.chief_ray = cr_pkg
        fld.ref_sphere = rs_pkg
        img_pt = cr_pkg[0].ray[-1][0]

        wvls = osp.spectral_region
        fans_x = []
        fans_y = []
        fan_start = np.array([0., 0.])
        fan_stop = np.array([0., 0.])
        fan_start[xy] = -1.0
        fan_stop[xy] = 1.0
        fan_def = [fan_start, fan_stop, num_rays]
        max_y_val = 0.0
        rc = []
        for wi, wvl in enumerate(wvls.wavelengths):
            rc.append(wvls.render_colors[wi])

            rs_pkg, cr_pkg = trace.setup_pupil_coords(self.opt_model,
                                                      fld, wvl, foc,
                                                      image_pt=img_pt)
            fld.chief_ray = cr_pkg
            fld.ref_sphere = rs_pkg
            fan = trace.trace_fan(self.opt_model, fan_def, fld, wvl, foc,
                                  img_filter=lambda p, ray_pkg:
                                  fct(p, xy, ray_pkg, fld, wvl, foc))
            f_x = []
            f_y = []
            for p, y_val in fan:
                f_x.append(p[xy])
                f_y.append(y_val)
                if abs(y_val) > max_y_val:
                    max_y_val = abs(y_val)
            fans_x.append(f_x)
            fans_y.append(f_y)
        fans_x = np.array(fans_x)
        fans_y = np.array(fans_y)
        return fans_x, fans_y, max_y_val, rc

    def trace_grid(self, fct, fi, wl=None, num_rays=21, form='grid',
                   append_if_none=True):
        """ fct is applied to the raw grid and returned as a grid  """
        osp = self.opt_model.optical_spec
        wvls = osp.spectral_region
        wvl = self.central_wavelength()
        wv_list = wvls.wavelengths if wl is None else [wvl]
        fld = osp.field_of_view.fields[fi]
        foc = osp.defocus

        rs_pkg, cr_pkg = trace.setup_pupil_coords(self.opt_model,
                                                  fld, wvl, foc)
        fld.chief_ray = cr_pkg
        fld.ref_sphere = rs_pkg

        grids = []
        grid_start = np.array([-1., -1.])
        grid_stop = np.array([1., 1.])
        grid_def = [grid_start, grid_stop, num_rays]
        for wi, wvl in enumerate(wv_list):
            grid = trace.trace_grid(self.opt_model, grid_def, fld, wvl, foc,
                                    form=form, append_if_none=append_if_none,
                                    img_filter=lambda p, ray_pkg:
                                    fct(p, wi, ray_pkg, fld, wvl, foc))
            grids.append(grid)
        rc = wvls.render_colors
        return grids, rc

    def trace_wavefront(self, fld, wvl, foc, num_rays=32):

        def wave(p, ray_pkg, fld, wvl, foc):
            x = p[0]
            y = p[1]
            if ray_pkg is not None:
                opd_pkg = rt.wave_abr(fld, wvl, foc, ray_pkg)
                opd = opd_pkg[0]/self.opt_model.nm_to_sys_units(wvl)
            else:
                opd = 0.0
            return np.array([x, y, opd])

        rs_pkg, cr_pkg = trace.setup_pupil_coords(self.opt_model,
                                                  fld, wvl, foc)
        fld.chief_ray = cr_pkg
        fld.ref_sphere = rs_pkg

        grid_start = np.array([-1., -1.])
        grid_stop = np.array([1., 1.])
        grid_def = (grid_start, grid_stop, num_rays)

        grid = trace.trace_grid(self.opt_model, grid_def, fld, wvl, foc,
                                img_filter=lambda p, ray_pkg:
                                wave(p, ray_pkg, fld, wvl, foc), form='grid')
        return grid

    # def set_clear_apertures(self):
    #     def rd(v):
    #         """ take 2d length of input vector v """
    #         return np.sqrt(v[0]*v[0]+v[1]*v[1])

    #     if self.get_num_surfaces() > 2:
    #         fields_df = trace.trace_all_fields(self.opt_model)
    #         # a) Select the inc_pt data from the unstacked result and
    #         #    transpose so that intrfcs are the index
    #         inc_pts = fields_df.unstack()['inc_pt'].T
    #         # b) applymap() is used to apply the function rd() to each
    #         #    element in the dataframe
    #         inc_pts_rd = inc_pts.applymap(rd)
    #         # c) apply max() function to each row (i.e. across columns,
    #         #    axis=1)
    #         semi_ap = inc_pts_rd.max(axis=1)
    #         for s, max_ap in zip(self.ifcs[1:-1], semi_ap[1:-1]):
    #             s.set_max_aperture(max_ap)

    def set_clear_apertures(self):
        rayset = trace.trace_boundary_rays(self.opt_model,
                                           use_named_tuples=True)

        for i, s in enumerate(self.ifcs):
            max_ap = -1.0e+10
            update = True
            for f in rayset:
                for p in f:
                    ray = p.ray
                    if len(ray) > i:
                        ap = sqrt(ray[i].p[0]**2 + ray[i].p[1]**2)
                        if ap > max_ap:
                            max_ap = ap
                    else:  # ray failed before this interface, don't update
                        update = False
            if update:
                s.set_max_aperture(max_ap)

    def trace(self, pt0, dir0, wvl, **kwargs):
        return rt.trace(self, pt0, dir0, wvl, **kwargs)

    def compute_global_coords(self, glo=1):
        """ Return global surface coordinates (rot, t) wrt surface glo. """
        tfrms = []
        r, t = np.identity(3), np.array([0., 0., 0.])
        prev = r, t
        tfrms.append(prev)
#        print(glo, t, *np.rad2deg(t3d.euler.mat2euler(r)))
        if glo > 0:
            # iterate in reverse over the segments before the
            #  global reference surface
            go = glo
            path = itertools.zip_longest(self.ifcs[glo::-1],
                                         self.gaps[glo-1::-1])
            after = next(path)
            # loop of remaining surfaces in path
            while True:
                try:
                    before = next(path)
                    go -= 1
                    zdist = after[Gap].thi
                    r, t = trns.reverse_transform(before[Intfc], zdist,
                                                  after[Intfc])
                    t = prev[0].dot(t) + prev[1]
                    r = prev[0].dot(r)
#                    print(go, t,
#                          *np.rad2deg(euler2opt(t3d.euler.mat2euler(r))))
                    prev = r, t
                    tfrms.append(prev)
                    after = before
                except StopIteration:
                    break
            tfrms.reverse()
        path = itertools.zip_longest(self.ifcs[glo:], self.gaps[glo:])
        before = next(path)
        prev = np.identity(3), np.array([0., 0., 0.])
        go = glo
        # loop forward over the remaining surfaces in path
        while True:
            try:
                after = next(path)
                go += 1
                zdist = before[Gap].thi
                r, t = trns.forward_transform(before[Intfc], zdist,
                                              after[Intfc])
                t = prev[0].dot(t) + prev[1]
                r = prev[0].dot(r)
#                print(go, t,
#                      *np.rad2deg(euler2opt(t3d.euler.mat2euler(r))))
                prev = r, t
                tfrms.append(prev)
                before = after
            except StopIteration:
                break
        return tfrms

    def compute_local_transforms(self):
        """ Return forward surface coordinates (r.T, t) for each interface. """
        tfrms = []
        path = itertools.zip_longest(self.ifcs, self.gaps)
        before = next(path)
        while before is not None:
            try:
                after = next(path)
            except StopIteration:
                tfrms.append((np.identity(3), np.array([0., 0., 0.])))
                break
            else:
                zdist = before[Gap].thi
                r, t = trns.forward_transform(before[Intfc], zdist,
                                              after[Intfc])
                rt = r.transpose()
                tfrms.append((rt, t))
                before = after

        return tfrms


def gen_sequence(surf_data_list, **kwargs):
    """ create a sequence iterator from the surf_data_list

    Args:
        surf_data_list: a list of lists containing:
                        [curvature, thickness, refractive_index, v-number]
        **kwargs: keyword arguments

    Returns:
        (**ifcs**, **gaps**, **rndx**, **lcl_tfrms**, **z_dir**)
    """
    ifcs = []
    gaps = []
    rndx = []
    lcl_tfrms = []
    z_dir = []

    for surf_data in surf_data_list:
        s, g, rn, tfrm = create_surface_and_gap(surf_data, **kwargs)
        ifcs.append(s)
        gaps.append(g)
        rndx.append(rn)
        lcl_tfrms.append(tfrm)
        z_dir.append(1)

    n_before = 1.0
    z_dir_before = 1
    for i, s in enumerate(ifcs):
        z_dir_after = copysign(1, z_dir_before)
        n_after = np.copysign(rndx[i], n_before)
        if s.interact_mode == 'reflect':
            n_after = -n_after
            z_dir_after = -z_dir_after

        n_before = n_after
        rndx[i] = n_after
        z_dir_before = z_dir_after
        z_dir[i] = z_dir_after

    seq = itertools.zip_longest(ifcs, gaps[:-2], lcl_tfrms, rndx, z_dir)
    return seq


def create_surface_and_gap(surf_data, radius_mode=False, prev_medium=None,
                           wvl=550.0, **kwargs):
    """ create a surface and gap where surf_data is a list that contains:
        [curvature, thickness, refractive_index, v-number] """
    s = surface.Surface()

    if radius_mode:
        if surf_data[0] != 0.0:
            s.profile.cv = 1.0/surf_data[0]
        else:
            s.profile.cv = 0.0
    else:
        s.profile.cv = surf_data[0]

    if len(surf_data) > 2:
        if isanumber(surf_data[2]):
            if len(surf_data) < 3:
                if surf_data[2] == 1.0:
                    mat = m.Air()
                else:
                    mat = m.Medium(surf_data[2])
            else:
                mat = m.Glass(surf_data[2], surf_data[3], '')

        else:
            if surf_data[2].upper() == 'REFL':
                s.interact_mode = 'reflect'
                mat = prev_medium
            else:
                name, cat = surf_data[2], surf_data[3]
                try:
                    mat = gfact.create_glass(name, cat)
                except ge.GlassNotFoundError as gerr:
                    logging.info('%s glass data type %s not found',
                                 gerr.catalog,
                                 gerr.name)
                    logging.info('Replacing material with air.')
                    mat = m.Air()

    else:  # only curvature and thickness entered, set material to air
        mat = m.Air()

    thi = surf_data[1]
    g = gap.Gap(thi, mat)
    rndx = mat.rindex(wvl)
    tfrm = np.identity(3), np.array([0., 0., thi])

    return s, g, rndx, tfrm



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" convert RGB data to matplotlib format

.. Created on Mon Jul 15 12:24:33 2019

.. codeauthor: Michael J. Hayford
"""


def rgb2mpl(rgb):
    """ convert 8 bit RGB data to 0 to 1 range for mpl """
    if len(rgb) == 3:
        return [rgb[0]/255., rgb[1]/255., rgb[2]/255.]
    elif len(rgb) == 4:
        return [rgb[0]/255., rgb[1]/255., rgb[2]/255., rgb[3]/255.]


backgrnd_color = rgb2mpl([237, 243, 254])  # light blue



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Lightweight manager class to connect a model+actions to windows

.. Created on Fri Aug 17 11:35:01 2018

.. codeauthor: Michael J. Hayford
"""
import logging

from collections import namedtuple

ModelInfo = namedtuple('ModelInfo', ['model', 'fct', 'args', 'kwargs'])
""" package of a model and a update function with args and kwargs

    Attributes:
        model: object associated with view update function
        fct: view update function, can be None
        args: list of fct arguments
        kwargs: list of fct keyword arguments

    Note:
        ``model`` is required; ``fct``, ``args`` and ``kwargs`` are optional
"""
ModelInfo.__new__.__defaults__ = (None, (), {})


class AppManager:
    """ Lightweight model/view manager class

    Lightweight manager class to manage connections between a model and
    a ui view.

    The main function of AppManager is refresh_gui(). This is called after
    a user input to the gui to update the model and call a refresh function
    for each ui view of that model.

    Attributes:
        model: the model of the currently active/frontmost ui view

            model is expected to respond to:

                - update_model()
                - name()

        gui_parent: the top level gui manager (optional)

            gui_parent is expected to implement:

                - refresh_app_ui()

        view_dict: keys are ui views, values are ModelInfo tuples

            view is expected to implement:

                - windowTitle() - only for debug logging

    """

    def __init__(self, model, gui_parent=None):
        self.model = self._tag_model(model)
        self.gui_parent = gui_parent
        self.view_dict = {}

    def _tag_model(self, model):
        if model is not None:
            if not hasattr(model, 'app_manager'):
                model.app_manager = self

    def set_model(self, model):
        self._tag_model(model)
        self.model = model

    def add_view(self, view, gui_hook, model_info):
        """ Add a new view and model tuple into dictionary

        Args:
            view: the ui view, used as a key
            gui_hook: instance of the GUI component to be refreshed
            model_info: instance of the ModelInfo tuple

        Returns:
            returns the input view
        """
        self._tag_model(model_info.model)
        self.view_dict[view] = gui_hook, model_info
        return view

    def delete_view(self, view):
        """ removes view from the view dictionary

        Args:
            view: view being closed by user
        """
        logging.debug('AppManager.delete_view: {}'.format(view.windowTitle()))
        del self.view_dict[view]

    def close_model(self, view_close_fct=None):
        """ close all ui views associated with the active model

        Args:
            view_close_fct: optional fct called on each closing view, with
                the view as an argument. This function, if used, should call
                delete_view itself.
        """
        cur_model = self.model
        view_items = list(self.view_dict.items())
        for view, info in view_items:
            mi = info[1]
            if mi.model == cur_model:
                if view_close_fct is None:
                    self.delete_view(view)
                else:
                    view_close_fct(view)
        delattr(cur_model, 'app_manager')
        self.model = None

    def refresh_gui(self):
        """ update the active model and refresh its dependent ui views """
        self.model.update_model()
        self.refresh_views()

    def refresh_views(self):
        """ refresh the dependent ui views of the active model """
        if self.gui_parent is not None:
            self.gui_parent.refresh_app_ui()
        # traverse a copy of the view dict. this way we can delete any errant
        #  views we might find. we don't seem to be trapping closing a window
        #  via a close box under pyqt5
        for view, info in dict(self.view_dict).items():
            mi = info[1]
            if mi.model == self.model:
                if mi.fct is not None:
                    try:
                        mi.fct(*mi.args, **mi.kwargs)
                    except RuntimeError:
                        del self.view_dict[view]

    def on_view_activated(self, view):
        """ Makes the model associated with input view the active model

        Args:
            view: view becoming the active view
        """
        if view is not None:
            try:
                info = self.view_dict[view]
            except KeyError:
                logging.debug('view "%s" not in view_dict',
                              view.windowTitle())
            else:
                mi = info[1]
                model = mi.model
                if model:
                    logging.debug("on_view_activated: %s, %s" %
                                  (model.name(), view.windowTitle()))
                if model and model != self.model:
                    self.model = model
                    self.refresh_views()
                    logging.debug("switch model to", model.name())

    def sync_light_or_dark(self, is_dark):
        """ Tells views to update to a light or dark color scheme """
        for view, _ in dict(self.view_dict).items():
            try:
                view.sync_light_or_dark(is_dark)
            except AttributeError:
                pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Interactive 2D lens picture

    The Lens Layout capability provides a 2D display of the optical model
    represented by **shapes**. Shapes contain `dict` attributes to manage the
    graphical rendering and editing actions associated with their parent
    objects. These attributes must include:

    **handles**: named graphic handles for different aspects of the parent object

    **actions**: functions for **press**, **drag**, and **release** actions

.. Created on Tue Sep 18 14:23:28 2018

.. codeauthor: Michael J. Hayford
"""

import numpy as np

from rayoptics.gui.util import (GUIHandle, transform_ray_seg, bbox_from_poly,
                                transform_poly, inv_transform_poly)

from rayoptics.optical import transform
from rayoptics.optical.trace import (trace_boundary_rays_at_field,
                                     boundary_ray_dict, RaySeg)
from rayoptics.optical.elements import (create_thinlens, create_mirror,
                                        create_lens, AirGap)
import rayoptics.optical.model_constants as mc
from rayoptics.util.rgb2mpl import rgb2mpl
import rayoptics.gui.appcmds as cmds


def setup_shift_of_ray_bundle(seq_model, start_offset):
    """ compute transformation for rays "start_offset" from 1st surface

    Args:
        seq_model: the sequential model
        start_offset: z distance rays should start wrt first surface.
                      positive if to left of first surface
    Returns:
        transformation rotation and translation::
            (rot, t)
    """

    s1 = seq_model.ifcs[1]
    s0 = seq_model.ifcs[0]
    rot, t = transform.reverse_transform(s0, start_offset, s1)
    return rot, t


def shift_start_of_ray_bundle(start_bundle, ray_bundle, rot, t):
    """ modify ray_bundle so that rays begin "start_offset" from 1st surface

    Args:
        ray_bundle: list of rays in a bundle, i.e. all for one field.
                    ray_bundle[0] is assumed to be the chief ray
        start_offset: z distance rays should start wrt first surface.
                      positive if to left of first surface
        rot: transformation rotation
        t: transformation translation
    """

    for ri, r in enumerate(ray_bundle):
        b4_pt = rot.dot(r.ray[1].p) - t
        b4_dir = rot.dot(r.ray[0].d)
        if ri == 0:
            # For the chief ray, use the input offset.
            dst = -b4_pt[2]/b4_dir[2]
        else:
            pt0 = start_bundle[0].p
            dir0 = start_bundle[0].d
            # Calculate distance along ray to plane perpendicular to
            #  the chief ray.
            dst = -(b4_pt - pt0).dot(dir0)/b4_dir.dot(dir0)
        pt = b4_pt + dst*b4_dir
        ray0 = RaySeg(pt, b4_dir, dst, r.ray[0].nrml)
        start_bundle[ri] = ray0


def create_optical_element(opt_model, e):
    return OpticalElement(opt_model, e)


class OpticalElement():
    """ mediator class for rendering and editing optical elements """

    def __init__(self, opt_model, e):
        self.opt_model = opt_model
        self.e = e
        self.select_pt = None
        self.move_direction = None
        self.handles = {}
        self.actions = self.edit_shape_actions()
        self.handle_actions = self.e.handle_actions()

    def update_shape(self, view):
        self.e.render_handles(self.opt_model)
        for key, graphics_handle in self.e.handles.items():
            poly = np.array(graphics_handle.polydata)
            poly_gbl, bbox = transform_poly(graphics_handle.tfrm, poly)
            if graphics_handle.polytype == 'polygon':
                p = view.create_polygon(poly_gbl, self.render_color(),
                                        zorder=2.5)
            elif graphics_handle.polytype == 'polyline':
                priority = 2.
                hc = 'red'
                if key == 'ct':
                    priority = 3.
                    hc = 'blue'
                p = view.create_polyline(poly_gbl, hilite=hc, zorder=priority)
            else:
                break
            self.handles[key] = GUIHandle(p, bbox)
        return self.handles

    def render_color(self):
        return self.e.render_color

    def get_label(self):
        if not hasattr(self.e, 'label'):
            self.e.label = 'element'
        return self.e.label

    def edit_shape_actions(self):
        def add_event_data(self, event, handle):
            gbl_pt = np.array([event.xdata, event.ydata])
            lcl_pt = inv_transform_poly(self.e.handles[handle].tfrm, gbl_pt)
            event.lcl_pt = lcl_pt
            if self.select_pt is not None:
                xdata, ydata = self.select_pt[1]
            else:
                xdata, ydata = 0., 0.
            dxdy = event.xdata - xdata, event.ydata - ydata
            event.dxdy = dxdy

        def on_select_shape(fig, handle, event, info):
            handle_actions = self.handle_actions[handle]
            add_event_data(self, event, handle)
            for key, action_obj in handle_actions.items():
                action_obj.actions['press'](fig, event)
            self.select_pt = ((event.x, event.y), (event.xdata, event.ydata))
#            print('select pt:', self.select_pt)
#            print('select pt:', event.x, event.y)

        def on_edit_shape(fig, handle, event, info):
            handle_actions = self.handle_actions[handle]
            x, y = self.select_pt[0]
            xdata, ydata = self.select_pt[1]
            delta_x, delta_y = abs(x - event.x), abs(y - event.y)
            delta_xdata, delta_ydata = (abs(xdata - event.xdata),
                                        abs(ydata - event.ydata))
            if self.move_direction is None:
                if delta_x > delta_y:
                    self.move_direction = 'x'
#                    print('move horizontal: delta x, y:', delta_x, delta_y,
#                          delta_xdata, delta_ydata)
                elif delta_x < delta_y:
                    self.move_direction = 'y'
#                    print('move vertical: delta x, y:', delta_x, delta_y,
#                          delta_xdata, delta_ydata)
                else:
                    self.move_direction = None
#                    print('move same: delta x, y:', delta_x, delta_y)

            add_event_data(self, event, handle)
#            print('move pt:', event.xdata, event.ydata, event.lcl_pt)
            if 'pt' in handle_actions:
                if 'drag' in handle_actions['pt'].actions:
                    handle_actions['pt'].actions['drag'](fig, event,
                                                         event.lcl_pt)
            elif self.move_direction in handle_actions:
                if 'drag' in handle_actions[self.move_direction].actions:
                    if self.move_direction == 'x':
                        handle_actions['x'].actions['drag'](fig, event,
                                                            event.dxdy[0])
                    elif self.move_direction == 'y':
                        handle_actions['y'].actions['drag'](fig, event,
                                                            event.dxdy[1])

            fig.refresh_gui()

        def on_release_shape(fig, handle, event, info):
            # print('release pt:', event.x, event.y)
            handle_actions = self.handle_actions[handle]
            add_event_data(self, event, handle)
            for key, action_obj in handle_actions.items():
                action_obj.actions['release'](fig, event)
            self.select_pt = None
            self.move_direction = None
            fig.refresh_gui()

        actions = {}
        actions['press'] = on_select_shape
        actions['drag'] = on_edit_shape
        actions['release'] = on_release_shape
        return actions


class RayBundle():
    """ class for ray bundle from a single field point """

    def __init__(self, opt_model, fld, fld_label, wvl, start_offset,
                 ray_table_callback=None):
        self.opt_model = opt_model
        self.fld = fld
        self.fld_label = fld_label
        self.wvl = wvl
        self.start_offset = start_offset
        self.handles = {}
        self.actions = self.edit_ray_bundle_actions()
        self.ray_table_callback = ray_table_callback

    def get_label(self):
        return self.fld_label

    def render_ray(self, ray, start_seg, tfrms):
        poly = []
        transform_ray_seg(poly, start_seg, tfrms[0])
        for i, r in enumerate(ray[1:], 1):
            transform_ray_seg(poly, r, tfrms[i])
        return np.array(poly)

    def render_shape(self, rayset, start_bundle, tfrms):
        poly1 = []
        transform_ray_seg(poly1, start_bundle[3], tfrms[0])
        for i, r in enumerate(rayset['+Y'].ray[1:], 1):
            transform_ray_seg(poly1, r, tfrms[i])

        poly2 = []
        transform_ray_seg(poly2, start_bundle[4], tfrms[0])
        for i, r in enumerate(rayset['-Y'].ray[1:], 1):
            transform_ray_seg(poly2, r, tfrms[i])

        poly2.reverse()
        poly1.extend(poly2)
        bbox = bbox_from_poly(poly1)
        return poly1, bbox

    def update_shape(self, view):
        rndr_clr = [254, 197, 254, 64]  # magenta, 25%
        wvl = self.opt_model.optical_spec.spectral_region.central_wvl
        self.wvl = wvl
        rayset = trace_boundary_rays_at_field(self.opt_model,
                                              self.fld, wvl,
                                              use_named_tuples=True)

        self.rayset = boundary_ray_dict(self.opt_model, rayset)
        # If the object distance (tfrms[0][1][2]) is greater than the
        #  start_offset, then modify rayset start to match start_offset.
        # Remember object transformation for resetting at the end.
        seq_model = self.opt_model.seq_model
        tfrms = seq_model.gbl_tfrms
        tfrtm0 = tfrms[0]

        start_bundle = [r.ray[0] for r in self.rayset.values()]
        if abs(tfrtm0[1][2]) > self.start_offset:
            rot, t = setup_shift_of_ray_bundle(seq_model, self.start_offset)
            tfrms[0] = (rot, t)
            shift_start_of_ray_bundle(start_bundle, self.rayset.values(),
                                      rot, t)

        try:
            poly, bbox = self.render_shape(self.rayset,
                                           start_bundle, tfrms)
            cr = self.render_ray(self.rayset['00'].ray,
                                 start_bundle[0], tfrms)
            upr = self.render_ray(self.rayset['+Y'].ray,
                                  start_bundle[3], tfrms)
            lwr = self.render_ray(self.rayset['-Y'].ray,
                                  start_bundle[4], tfrms)
        finally:
            tfrms[0] = tfrtm0

        p = view.create_polygon(poly, rndr_clr)
        self.handles['shape'] = GUIHandle(p, bbox)

        cr_poly = view.create_polyline(cr)
        self.handles['00'] = GUIHandle(cr_poly, bbox_from_poly(cr))

        upr_poly = view.create_polyline(upr)
        self.handles['+Y'] = GUIHandle(upr_poly, bbox_from_poly(upr))

        lwr_poly = view.create_polyline(lwr)
        self.handles['-Y'] = GUIHandle(lwr_poly, bbox_from_poly(lwr))

        return self.handles

    def edit_ray_bundle_actions(self):
        def on_select_ray(fig, handle, event, info):
            if handle != 'shape':
                ray_table = self.ray_table_callback()
                ray_table.root = self.rayset[handle].ray
                fig.refresh_gui()

        actions = {}
        actions['press'] = on_select_ray
        return actions


class ParaxialRay():
    """ class for paraxial ray rendering/editing """

    def __init__(self, opt_model, ray, color, seq_start=1, label='paraxial'):
        self.label = label
        self.opt_model = opt_model
        self.ray = ray
        self.seq_start = seq_start
        self.color = color
        self.handles = {}
        self.actions = self.edit_paraxial_layout_actions()
        self.vertex = None

    def get_label(self):
        return self.label

    def render_ray(self, ray, tfrms):
        poly = []
        for i, r in enumerate(ray[self.seq_start:], self.seq_start):
            rot, trns = tfrms[i]
            ps = np.array([0., r[mc.ht], 0.])
            p = rot.dot(ps) + trns
            poly.append([p[2], p[1]])
        return np.array(poly)

    def update_shape(self, view):
        seq_model = self.opt_model.seq_model
        tfrms = seq_model.gbl_tfrms

        ray_poly = self.render_ray(self.ray, tfrms)

        priority = 3
        rp = view.create_polyline(ray_poly, color=self.color, linewidth=1,
                                  hilite=self.color, zorder=priority)
        self.handles['shape'] = GUIHandle(rp, bbox_from_poly(ray_poly))

        return self.handles

    def apply_data(self, vertex, lcl_pt):
        ray = self.ray
        p = ray[vertex-1]
        c = ray[vertex]
        c_slp_new = (lcl_pt[1] - c[mc.ht])/lcl_pt[0]
        pwr = (p[mc.slp] - c_slp_new)/c[mc.ht]
        self.opt_model.seq_model.ifcs[vertex].optical_power = pwr

    def edit_paraxial_layout_actions(self):
        def add_event_data(shape, event, handle, info):
            gbl_pt = np.array([event.xdata, event.ydata])
            lcl_pt = inv_transform_poly(shape.tfrm, gbl_pt)
            event.lcl_pt = lcl_pt

        def on_select_point(fig, handle, event, info):
            self.vertex = self.seq_start
            if 'ind' in info:
                self.vertex += info['ind'][0]
            seq_model = self.opt_model.seq_model
            self.tfrm = seq_model.gbl_tfrms[self.vertex]

        def on_edit_point(fig, handle, event, info):
            add_event_data(self, event, handle, info)
            self.apply_data(self.vertex, event.lcl_pt)
            fig.refresh_gui()

        def on_release_point(fig, handle, event, info):
            add_event_data(self, event, handle, info)
            self.apply_data(self.vertex, event.lcl_pt)
            fig.refresh_gui()
            self.vertex = None
            self.tfrm = None

        actions = {}
        actions['press'] = on_select_point
        actions['drag'] = on_edit_point
        actions['release'] = on_release_point
        return actions


class LensLayout():
    """ manager for live layout graphics entities """

    def __init__(self, opt_model, **kwargs):
        self.opt_model = opt_model
        self.ray_table = None

        ele_model = self.opt_model.ele_model
        if len(ele_model.elements) == 0:
            ele_model.elements_from_sequence(self.opt_model.seq_model)

    def system_length(self, ele_bbox, offset_factor=0.05):
        """ returns system length and ray start offset """
        specsheet = self.opt_model.specsheet
        osp = self.opt_model.optical_spec
        if specsheet.conjugate_type == 'finite':
            return specsheet.imager.tt, (2/3)*specsheet.imager.sp
        elif specsheet.conjugate_type == 'infinite':
            return specsheet.imager.sp, offset_factor*specsheet.imager.sp
        else:
            img_dist = abs(osp.parax_data[2].img_dist)
            ele_length = ele_bbox[1][0] - ele_bbox[0][0]
            return ele_length+img_dist, offset_factor*img_dist

    def create_element_model(self, view):
        ele_model = self.opt_model.ele_model
        elements = [create_optical_element(self.opt_model, e)
                    for e in ele_model.elements]
        return elements

    def create_ray_model(self, view, start_offset):
        ray_bundles = []
        fov = self.opt_model.optical_spec.field_of_view
        wvl = self.opt_model.seq_model.central_wavelength()
        for i, fld in enumerate(fov.fields):
            fld_label = fov.index_labels[i]
            rb = RayBundle(self.opt_model, fld, fld_label, wvl, start_offset,
                           ray_table_callback=self.get_ray_table)
            ray_bundles.append(rb)
        return ray_bundles

    def create_paraxial_layout(self, view):
        parax_model = self.opt_model.parax_model

        ax_color = rgb2mpl([138, 43, 226])  # blueviolet
        ax_poly = ParaxialRay(self.opt_model, parax_model.ax, color=ax_color,
                              label='axial ray')
        pr_color = rgb2mpl([198, 113, 113])  # sgi salmon
        pr_poly = ParaxialRay(self.opt_model, parax_model.pr, color=pr_color,
                              label='chief ray')
        return [ax_poly, pr_poly]

    def get_ray_table(self):
        if self.ray_table is None:
            self.ray_table = cmds.create_ray_table_model(self.opt_model, None)
            gui_parent = self.opt_model.app_manager.gui_parent
            gui_parent.create_table_view(self.ray_table, "Ray Table")

        return self.ray_table

    def register_commands(self, *args, **kwargs):
        self.apply_fct = kwargs.pop('apply_fct')
        fig = kwargs.pop('figure')
        actions = self.add_element_cmd_actions(**kwargs)

        def do_command_action(event, target, event_key):
            nonlocal fig, actions
            if target is not None:
                shape, handle = target.artist.shape
                if handle == 'ct' and isinstance(shape.e, AirGap):
                    try:
                        action = actions[event_key]
                        action(fig, shape, event, target.info)
                    except KeyError:
                        pass
        fig.do_action = do_command_action

    def add_element_cmd_actions(self, **kwargs):
        idx = None

        def add_event_data(tfrm, event):
            gbl_pt = np.array([event.xdata, event.ydata])
            lcl_pt = inv_transform_poly(tfrm, gbl_pt)
            event.lcl_pt = lcl_pt

        def on_select_cmd(fig, shape, event, info):
            nonlocal idx
            idx = shape.e.idx
            tfrm = self.opt_model.seq_model.gbl_tfrms[idx]
            add_event_data(tfrm, event)
            self.apply_fct(self.opt_model, idx, event.lcl_pt, **kwargs)
            fig.refresh_gui()

#        def on_edit_cmd(fig, shape, event, info):
#            add_event_data(self, event)
#            self.apply_fct(idx, event.lcl_pt)
#            fig.refresh_gui()
#        actions['drag'] = on_edit_cmd

        def on_release_cmd(fig, shape, event, info):
            nonlocal idx
#            add_event_data(self, event)
#            self.apply_fct(idx, event.lcl_pt)
#            fig.refresh_gui()
            idx = None

        actions = {}
        actions['press'] = on_select_cmd
        actions['release'] = on_release_cmd
        return actions


def split_gap(opt_model, idx, lcl_pt):
    """ split g=gap[idx] into t_old = t_0 + t_k using t_0 = lcl_pt.x """
    g = opt_model.seq_model.gaps[idx]
    x, y = lcl_pt
    t_k = g.thi - x
    t_0 = x
    return g, t_0, t_k


def add_elements(opt_model, idx, lcl_pt, create, **kwargs):
    g, t_0, t_k = split_gap(opt_model, idx, lcl_pt)
    g.thi = t_0
    opt_model.insert_ifc_gp_ele(*create(**kwargs), idx=idx, t=t_k,
                                insert=True)


def add_reflector(opt_model, idx, lcl_pt, create, **kwargs):
    g, t_0, t_k = split_gap(opt_model, idx, lcl_pt)
    g.thi = t_0
    opt_model.insert_ifc_gp_ele(*create(**kwargs), idx=idx, t=-t_k,
                                insert=True)


def add_thinlens(opt_model, idx, lcl_pt, **kwargs):
    add_elements(opt_model, idx, lcl_pt, create_thinlens, **kwargs)


def add_lens(opt_model, idx, lcl_pt, **kwargs):
    g, t_0, t_k = split_gap(opt_model, idx, lcl_pt)
    t_old = g.thi
    if 'th' in kwargs:
        t_ct = kwargs['th']
    else:
        t_ct = 0.05*t_old
    t0_new = t_0 - t_ct/2
    tk_new = t_k - t_ct/2
    g.thi = t0_new

    seq, ele = create_lens(th=t_ct)
    opt_model.insert_ifc_gp_ele(seq, ele, idx=idx, t=tk_new, insert=True)


def add_mirror(opt_model, idx, lcl_pt, **kwargs):
    add_reflector(opt_model, idx, lcl_pt, create_mirror, **kwargs)


def add_conic(opt_model, idx, lcl_pt, **kwargs):
    add_reflector(opt_model, idx, lcl_pt, create_mirror, **kwargs)


def add_doublet(opt_model, idx, lcl_pt, **kwargs):
    add_elements(opt_model, idx, lcl_pt, create_lens, **kwargs)



# -*- coding: utf-8 -*-
""" The **ray-optics** geometrical ray tracing and optical modeling and
    analysis package

    The basic optical functionality is contained in the :mod:`rayoptics.optical`
    subpackage. It is supported by the following subpackages:

        - :mod:`rayoptics.codev`: handles import of CODE V .seq files
        - :mod:`opticalglass`: this package interfaces with glass manufacturer optical data

    The :mod:`rayoptics.gui` subpackage is a layer that implements the platform
    neutral part of the graphical user interface.

    The :mod:`rayoptics.mpl` subpackage implements a variety of plotting/charting
    based graphics using the **matplotlib** package. The :mod:`rayoptics.qtgui`
    subpackage implements a desktop ui style application using **Qt** featuring
    interactive layouts and diagrams.

    The :mod:`rayoptics.util` subpackage provides a variety of different math
    and other miscellaneous calculations.
"""

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except:
    __version__ = 'unknown'

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Module for simple optical media definitions

.. Created on Fri Sep 15 17:06:17 2017

.. codeauthor: Michael J. Hayford
"""

from scipy.interpolate import interp1d

from rayoptics.util.spectral_lines import spectra


def glass_encode(n, v):
    return str(1000*round((n - 1), 3) + round(v/100, 3))


def glass_decode(gc):
    return round(1.0 + (int(gc)/1000), 3), round(100.0*(gc - int(gc)), 3)


class Medium:
    """ Constant refractive index medium. """
    def __init__(self, nd, lbl):
        self.label = lbl
        self.n = nd

    def __repr__(self):
        return 'Medium ' + self.label + ': ' + str(self.n)

    def name(self):
        return self.label

    def rindex(self, wv_nm):
        """ returns the interpolated refractive index at wv_nm

        Args:
            wv_nm: the wavelength in nm for the refractive index query

        Returns:
            float: the refractive index at wv_nm
        """
        return self.n


class Air(Medium):
    """ Optical definition for air (low fidelity definition) """
    def __init__(self):
        self.label = 'air'
        self.n = 1.0

    def __repr__(self):
        return 'Air'


class Glass(Medium):
    """ Optical medium defined by a glass code, i.e. index - V number pair """
    def __init__(self, nd=1.5168, vd=64.17, mat='N-BK7'):
        self.label = mat
        if mat == 'N-BK7':
            self.n = 1.5168
            self.v = 64.17
        else:
            self.n = nd
            self.v = vd

    def __repr__(self):
        return 'Glass ' + self.label + ': ' + glass_encode(self.n, self.v)

    def glass_code(self):
        return str(1000*round((self.n - 1), 3) + round(self.v/100, 3))

    def name(self):
        if self.label == '':
            return glass_encode(self.n, self.v)
        else:
            return self.label

    def rindex(self, wv_nm):
        return self.n


class InterpolatedGlass():
    """ Optical medium defined by a list of wavelength/index pairs

    Attributes:
        label: required string identifier for the material
        wvls: list of wavelenghts in nm, used as x axis
        rndx: list of refractive indices corresponding to the values in wvls
        rindex_interp: the interpolation function
    """
    def __init__(self, label, pairs=None, rndx=None, wvls=None):
        self.label = label
        if pairs is not None:
            self.wvls = []
            self.rndx = []
            for w, n in pairs:
                self.wvls.append(w)
                self.rndx.append(n)
        else:
            self.wvls = wvls
            self.rndx = rndx
        self.update()

    def __repr__(self):
        return ('InterpolatedGlass(' + repr(self.label) +
                ', wvls=' + repr(self.wvls) +
                ', rndx=' + repr(self.rndx) + ')')

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['rindex_interp']
        return attrs

    def sync_to_restore(self):
        """ rebuild interpolating function """
        self.update()

    def update(self):
        self.rindex_interp = interp1d(self.wvls, self.rndx, kind='cubic',
                                      assume_sorted=False)

    def glass_code(self):
        nd = self.rindex('d')
        nF = self.rindex('F')
        nC = self.rindex('C')
        vd = (nd - 1)/(nF - nC)
        return str(glass_encode(nd, vd))

    def name(self):
        if self.label == '':
            return self.glass_code()
        else:
            return self.label

    def rindex(self, wv_nm):
        """ returns the interpolated refractive index at wv_nm

        Args:
            wvl: either the wavelength in nm or a string with a spectral line
                 identifier. for the refractive index query

        Returns:
            float: the refractive index at wv_nm

        Raises:
            KeyError: if ``wvl`` is not in the spectra dictionary
        """
        if isinstance(wv_nm, float):
            return float(self.rindex_interp(wv_nm))
        elif isinstance(wv_nm, int):
            return float(self.rindex_interp(wv_nm))
        else:
            return float(self.rindex_interp(spectra[wv_nm]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Module for optical surface related classes

    Surface
        Container of profile, extent, position and orientation information of
        the surface

    DecenterData
        Maintains data and actions to support 4 types of position and
        orientation changes.

        - DEC: pos and orientation applied prior to surface
        - REV: pos and orientation applied following surface in reverse
        - DAR: pos and orientation applied prior to surface and then returned to initial frame
        - BEN: used for fold mirrors, orientation applied before and after surface

    Aperture
        - Circular
        - Rectangular
        - Elliptical

.. Created on Sat Sep 16 09:22:05 2017

.. codeauthor: Michael J. Hayford
"""

from enum import Enum, auto
from math import sqrt
import numpy as np

from . import profiles
import transforms3d as t3d
from .model_enums import DecenterType as dec
from rayoptics.optical.traceerror import TraceError


class InteractionMode(Enum):
    """ enum for different interact_mode specifications """
    Transmit = auto()  #: propagate in transmission at this interface
    Reflect = auto()   #: propagate in reflection at this interface


class Interface:
    def __init__(self, interact_mode='transmit', delta_n=0.0,
                 max_ap=1.0, decenter=None, phase_element=None):
        self.interact_mode = interact_mode
        self.delta_n = delta_n
        self.decenter = decenter
        self.max_aperture = max_ap
        if phase_element is not None:
            self.phase_element = phase_element

    def update(self):
        if self.decenter is not None:
            self.decenter.update()

    def interface_type(self):
        return type(self).__name__

    def sync_to_restore(self, opt_model):
        if not hasattr(self, 'max_aperture'):
            self.max_aperture = 1.0
        if hasattr(self, 'interact_mode'):
            if isinstance(self.interact_mode, InteractionMode):
                if self.interact_mode == InteractionMode.Reflect:
                    self.interact_mode = 'reflect'
                elif self.interact_mode == InteractionMode.Transmit:
                    self.interact_mode = 'transmit'
        if hasattr(self, 'refract_mode'):  # really old models
            if self.refract_mode == 'REFL':
                self.interact_mode = 'reflect'
            else:
                self.interact_mode = 'transmit'
            delattr(self, 'refract_mode')

    @property
    def profile_cv(self):
        return 0.0

    def set_optical_power(self, pwr, n_before, n_after):
        pass

    def surface_od(self):
        pass

    def set_max_aperture(self, max_ap):
        """ max_ap is the max aperture radius """
        self.max_aperture = max_ap

    def intersect(self, p0, d, eps=1.0e-12):
        pass

    def normal(self, p):
        pass

    def phase(self, pt, d_in, normal, wl):
        if hasattr(self, 'phase_element'):
            return self.phase_element.phase(pt, d_in, normal, wl=wl)

    def apply_scale_factor(self, scale_factor):
        self.max_aperture *= scale_factor
        if self.decenter:
            self.decenter.apply_scale_factor(scale_factor)


class Surface(Interface):
    """ Container of profile, extent, position and orientation. """

    def __init__(self, lbl='', profile=None, **kwargs):
        super().__init__(**kwargs)
        self.label = lbl
        if profile:
            self.profile = profile
        else:
            self.profile = profiles.Spherical()
        self.clear_apertures = []
        self.edge_apertures = []

    def __repr__(self):
        if len(self.label) > 0:
            return "{!s}(lbl={!r}, profile={!r}, interact_mode={!s})" \
                   .format(type(self).__name__,
                           self.label, self.profile, self.interact_mode)
        else:
            return "{!s}(profile={!r}, interact_mode={!s})" \
                   .format(type(self).__name__,
                           self.profile, self.interact_mode)

    def interface_type(self):
        return type(self.profile).__name__

    def update(self):
        super().update()
        self.profile.update()

    def sync_to_restore(self, opt_model):
        super().sync_to_restore(opt_model)
        for ca in self.clear_apertures:
            ca.sync_to_restore(opt_model)
        for ea in self.edge_apertures:
            ea.sync_to_restore(opt_model)

    @property
    def profile_cv(self):
        return self.profile.cv

    @profile_cv.setter
    def profile_cv(self, cv):
        self.profile.cv = cv

    @property
    def optical_power(self):
        return self.delta_n * self.profile.cv

    @optical_power.setter
    def optical_power(self, pwr):
        self.profile.cv = pwr/self.delta_n if self.delta_n != 0.0 else 0.0

    def set_optical_power(self, pwr, n_before, n_after):
        self.delta_n = n_after - n_before
        self.optical_power = pwr

    def apply_scale_factor(self, scale_factor):
        super().apply_scale_factor(scale_factor)
        self.max_aperture *= scale_factor
        self.profile.apply_scale_factor(scale_factor)
        for e in self.edge_apertures:
            e.apply_scale_factor(scale_factor)
        for ca in self.clear_apertures:
            ca.apply_scale_factor(scale_factor)

    def from_first_order(self, nu_before, nu_after, y):
        pass

    def z_sag(self, pt):
        return self.profile.sag(0., pt[1])

    def set_z_sag(self, pt):
        self.profile.cv = self.calc_cv_from_zsag(pt)

    def calc_cv_from_zsag(self, pt):
        x, y = pt
        cv = 2*x / (x**2 + y**2)
        return cv

    def surface_od(self):
        od = 0
        if len(self.edge_apertures) > 0:
            for e in self.edge_apertures:
                edg = e.max_dimension()
                if edg > od:
                    od = edg
        elif len(self.clear_apertures) > 0:
            for ca in self.clear_apertures:
                ap = ca.max_dimension()
                if ap > od:
                    od = ap
        else:
            od = self.max_aperture

        return od

    def get_y_aperture_extent(self):
        """ returns [y_min, y_max] for the union of apertures """
        od = [1.0e10, -1.0e10]
        if len(self.edge_apertures) > 0:
            for e in self.edge_apertures:
                edg = e.bounding_box()
                if edg[0][1] < od[0]:
                    od[0] = edg[0][1]
                if edg[1][1] > od[1]:
                    od[1] = edg[1][1]
        elif len(self.clear_apertures) > 0:
            for ca in self.clear_apertures:
                ap = ca.bounding_box()
                if ap[0][1] < od[0]:
                    od[0] = ap[0][1]
                if ap[1][1] > od[1]:
                    od[1] = ap[1][1]
        else:
            od = [-self.max_aperture, self.max_aperture]

        return od

    def full_profile(self, edge_extent, flat_id=None, dir=1, steps=6):
        if flat_id is None:
            return self.profile.profile(edge_extent, dir, steps)
        else:
            if len(edge_extent) == 1:
                sd_upr = edge_extent[0]
                sd_lwr = -edge_extent[0]
            else:
                sd_upr = edge_extent[1]
                sd_lwr = edge_extent[0]
            if dir < 0:
                sd_lwr, sd_upr = sd_upr, sd_lwr

            prf = []
            try:
                sag = self.profile.sag(0, flat_id)
            except TraceError:
                sag = None
            else:
                prf.append([sag, sd_lwr])
            prf += self.profile.profile((flat_id,), dir, steps)
            if sag is not None:
                prf.append([sag, sd_upr])
            return prf

    def intersect(self, p0, d, eps=1.0e-12, z_dir=1.0):
        return self.profile.intersect(p0, d, eps, z_dir)

    def normal(self, p):
        return self.profile.normal(p)


class DecenterData():
    """ Maintains data and actions for position and orientation changes.

        - LOCAL: pos and orientation applied prior to surface
        - REV:   pos and orientation applied following surface in reverse
        - DAR:   pos and orientation applied prior to surface and then returned to initial frame
        - BEND:  used for fold mirrors, orientation applied before and after surface

    """
    def __init__(self, dtype, x=0., y=0., alpha=0., beta=0., gamma=0.):
        self.dtype = dtype
        # x, y, z vertex decenter
        self.dec = np.array([x, y, 0.])
        # alpha, beta, gamma euler angles
        self.euler = np.array([alpha, beta, gamma])
        # x, y, z rotation point offset
        self.rot_pt = np.array([0., 0., 0.])
        self.rot_mat = None

    def __repr__(self):
        return "%r: Decenter: %r, Tilt: %r" % (self.dtype.name, self.dec,
                                               self.euler)

    def update(self):
        def convertl2r(self):
            return np.array([-self.euler[0], -self.euler[1], self.euler[2]])
        if self.euler.any():
            self.rot_mat = t3d.euler.euler2mat(*np.deg2rad(convertl2r(self)))
        else:
            self.rot_mat = None

    def apply_scale_factor(self, scale_factor):
        self.dec *= scale_factor
        self.rot_pt *= scale_factor

    def tform_before_surf(self):
        if self.dtype is not dec.REV:
            return self.rot_mat, self.dec
        else:
            return None, np.array([0., 0., 0.])

    def tform_after_surf(self):
        if self.dtype is dec.REV or self.dtype is dec.DAR:
            return self.rot_mat.transpose(), -self.dec
        elif self.dtype is dec.BEND:
            return self.rot_mat, np.array([0., 0., 0.])
        else:
            return None, np.array([0., 0., 0.])


class Aperture():
    def __init__(self, x_offset=0.0, y_offset=0.0, rotation=0.0):
        self.x_offset = x_offset
        self.y_offset = y_offset
        self.rotation = rotation

    def sync_to_restore(self, opt_model):
        if not hasattr(self, 'x_offset'):
            self.x_offset = 0.0
        if not hasattr(self, 'y_offset'):
            self.y_offset = 0.0
        if not hasattr(self, 'rotation'):
            self.rotation = 0.0

    def dimension(self):
        pass

    def set_dimension(self, x, y):
        pass

    def max_dimension(self):
        x, y = self.dimension()
        return sqrt(x*x + y*y)

    def bounding_box(self):
        center = np.array([self.x_offset, self.y_offset])
        extent = np.array(self.dimension())
        return center-extent, center+extent

    def apply_scale_factor(self, scale_factor):
        self.x_offset *= scale_factor
        self.y_offset *= scale_factor


class Circular(Aperture):
    def __init__(self, radius=1.0, **kwargs):
        super().__init__(**kwargs)
        self.radius = radius

    def dimension(self):
        return (self.radius, self.radius)

    def set_dimension(self, x, y):
        self.radius = sqrt(x*x + y*y)

    def max_dimension(self):
        return self.radius

    def apply_scale_factor(self, scale_factor):
        super().apply_scale_factor(scale_factor)
        self.radius *= scale_factor


class Rectangular(Aperture):
    def __init__(self, x_half_width=1.0, y_half_width=1.0, **kwargs):
        super().__init__(**kwargs)
        self.x_half_width = x_half_width
        self.y_half_width = y_half_width

    def dimension(self):
        return (self.x_half_width, self.y_half_width)

    def set_dimension(self, x, y):
        self.x_half_width = abs(x)
        self.y_half_width = abs(y)

    def apply_scale_factor(self, scale_factor):
        super().apply_scale_factor(scale_factor)
        self.x_half_width *= scale_factor
        self.y_half_width *= scale_factor


class Elliptical(Aperture):
    def __init__(self, x_half_width=1.0, y_half_width=1.0, **kwargs):
        super().__init__(**kwargs)
        self.x_half_width = x_half_width
        self.y_half_width = y_half_width

    def dimension(self):
        return (self.x_half_width, self.y_half_width)

    def set_dimension(self, x, y):
        self.x_half_width = abs(x)
        self.y_half_width = abs(y)

    def apply_scale_factor(self, scale_factor):
        super().apply_scale_factor(scale_factor)
        self.x_half_width *= scale_factor
        self.y_half_width *= scale_factor



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" module to setup an ideal imager

.. Created on Thu May 16 19:57:47 2019

.. codeauthor: Michael J. Hayford
"""

import math
from collections import namedtuple


ideal_imager_keys = ['m', 's', 'sp', 'tt', 'f']
ideal_imager_labels = ["m", "s", "s'", "tt", "f"]
IdealImager = namedtuple('IdealImager', ideal_imager_keys)
""" tuple grouping together first order specifications

    Attributes:
        m: (lateral) magnification
        s: object distance from first principal plane, P1->Obj
        sp: image distance from second principal plane, P2->Img
        tt: total track length, tt = sp - s
        f: focal length
"""


def ideal_imager_setup(**inputs):
    """ Calculate the ideal imaging properties given two independent parameters

    Given 2 system parameters from the following list, this function
    calculates the remaining parameters.

    Note that if specifying ``tt`` and ``f``, their ratio, tt/f, must be
    greater than or equal to 4. A `ValueError` is raised otherwise.

    For a typical system, the value of ``s`` is negative, i.e. the object is to
    the left of the first principal plane.

    Example::

        In [3]: m1s1 = ideal_imager_setup(m=-0.5, s=-10.0); m1s1
        Out[3]: IdealImager(m=-0.5, s=-10.0, sp=5.0, tt=15.0, f=3.333333333333)

        In [4]: s_inf_efl = ideal_imager_setup(s=-math.inf, f=25.0); s_inf_efl
        Out[4]: IdealImager(m=-0.0, s=-inf, sp=25.0, tt=inf, f=25.0)

    Args:
        m: (lateral) magnification
        s: object distance from first principal plane, P1->Obj
        sp: image distance from second principal plane, P2->Img
        tt: total track length, tt = sp - s
        f: focal length

    Returns:
        :class:`IdealImager` namedtuple

    Raises:
        ValueError: if tt/f < 4
    """
    if 'm' in inputs:
        m = inputs['m']
        if 's' in inputs:
            s = inputs['s']
            sp = m*s
            tt = sp - s
            f = s*sp/(s - sp)
        elif 'sp' in inputs:
            sp = inputs['sp']
            s = sp/m
            tt = sp - s
            f = s*sp/(s - sp)
        elif 'tt' in inputs:
            tt = inputs['tt']
            s = tt/(m - 1)
            sp = m*s
            f = s*sp/(s - sp)
        elif 'f' in inputs:
            f = inputs['f']
            tt = -f*(m - 1)**2/m
            s = -f*(m - 1)/m  # = tt/(m - 1)
            sp = m*s
        else:
            return IdealImager(m, None, None, None, None)

    elif 's' in inputs:
        # arrange calculations so that s=-inf is handled gracefully
        s = inputs['s']
        if 'sp' in inputs:
            sp = inputs['sp']
            f = 1/(1/sp - 1/s)
            m = sp/s
            tt = sp - s
        elif 'tt' in inputs:
            tt = inputs['tt']
            m = 1 + tt/s
            sp = m*s
            f = s*sp/(s - sp)
        elif 'f' in inputs:
            f = inputs['f']
            m = f/(s + f)
            sp = 1/(1/f + 1/s)
            tt = sp - s
        else:
            return IdealImager(None, s, None, None, None)

    elif 'sp' in inputs:
        # arrange calculations so that sp=inf is handled gracefully
        sp = inputs['sp']
        if 'tt' in inputs:
            tt = inputs['tt']
            m = sp/(sp - tt)
            s = sp/m
            f = s*sp/(s - sp)
        elif 'f' in inputs:
            f = inputs['f']
            m = (f - sp)/f
            s = 1/(1/sp - 1/f)
            tt = sp - s
        else:
            return IdealImager(None, None, sp, None, None)

    elif 'tt' in inputs:
        tt = inputs['tt']
        if 'f' in inputs:
            f = inputs['f']
            ttf = tt/f
            # tt/f >= 4, else no solution
            # pick root (+) that gives |s|>=|sp|, i.e. -1 <= m < 0
            m = ((2 - ttf) + math.sqrt(ttf*(ttf - 4)))/2
            s = tt/(m - 1)
            sp = m*s
        else:
            return IdealImager(None, None, None, tt, None)

    elif 'f' in inputs:
        f = inputs['f']
        return IdealImager(None, None, None, None, f)

    else:
        return IdealImager(None, None, None, None, None)

    return IdealImager(m, s, sp, tt, f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" generic ray optics commands for creating plots and tables

.. Created on Thu Nov  8 21:21:57 2018

.. codeauthor: Michael J. Hayford
"""

import os.path
import json_tricks
import rayoptics.codev.cmdproc as cvp

from rayoptics.optical.opticalmodel import OpticalModel
from rayoptics.optical.profiles import Spherical, Conic
from rayoptics.optical.elements import create_thinlens
from rayoptics.optical.firstorder import specsheet_from_parax_data
from rayoptics.optical.specsheet import (create_specsheets,
                                         create_specsheet_from_model)

from rayoptics.gui import layout
from rayoptics.gui.appmanager import ModelInfo
import rayoptics.gui.diagram as diagram

from rayoptics.mpl.lenslayoutfigure import LensLayoutFigure
from rayoptics.mpl.interactivelayout import InteractiveLayout
from rayoptics.mpl.axisarrayfigure import Fit
from rayoptics.mpl.axisarrayfigure import (RayFanFigure, SpotDiagramFigure,
                                           WavefrontFigure)

from rayoptics.mpl.analysisplots import FieldCurveFigure, ThirdOrderBarChart
from rayoptics.mpl.paraxdgnfigure import ParaxialDesignFigure
import rayoptics.mpl.interactivediagram as dgm

import rayoptics.qtgui.plotview as plotview
from rayoptics.qtgui.idealimagerdialog import IdealImagerDialog
from rayoptics.qtgui.pytablemodel import PyTableModel
from rayoptics.qtgui.plotview import (create_plot_scale_panel,
                                      create_draw_rays_groupbox,
                                      create_diagram_controls_groupbox,
                                      create_2d_figure_toolbar)


def open_model(file_name):
    """ open a file and populate an optical model with the data

    Args:
        file_name (str): a filename of a supported file type

            - .roa - a rayoptics JSON encoded file
            - .seq - a CODE V (TM) sequence file

    Returns:
        if successful, an OpticalModel instance, otherwise, None
    """
    file_extension = os.path.splitext(file_name)[1]
    opm = None
    if file_extension == '.seq':
        opm = cvp.read_lens(file_name)
        create_specsheet_from_model(opm)
    elif file_extension == '.roa':
        with open(file_name, 'r') as f:
            obj_dict = json_tricks.load(f)
            if 'optical_model' in obj_dict:
                opm = obj_dict['optical_model']
                opm.sync_to_restore()
    return opm


def create_new_model():
    return create_new_optical_system()


def create_new_optical_system(efl=10.0, fov=1.0):
    opt_model = OpticalModel()
    seq_model = opt_model.seq_model

    # put in minimum calculation defaults
    opt_model.seq_model.gaps[0].thi = 1.0e10
    opt_model.optical_spec.field_of_view.type = 'OBJ_ANG'
    opt_model.optical_spec.field_of_view.set_from_list([0., fov])

    opt_model.insert_ifc_gp_ele(*create_thinlens(power=1/efl, indx=1.5),
                                idx=0, t=efl, insert=True)

    opt_model.ele_model.add_dummy_interface_at_image(seq_model,
                                                     seq_model.gbl_tfrms)

    opt_model.update_model()

    return opt_model


def create_new_optical_model_from_specsheet(specsheet):
    """ create an OpticalModel with a basic thinlens model, given specsheet """
    opt_model = OpticalModel(specsheet=specsheet)

    seq_model = opt_model.seq_model

    # enter a basic thinlens model for the given specsheet
    imager = specsheet.imager
    if specsheet.conjugate_type == 'finite':
        opt_model.seq_model.gaps[0].thi = -imager.s
    else:
        opt_model.seq_model.gaps[0].thi = 1.0e10

    opt_model.insert_ifc_gp_ele(*create_thinlens(power=1/imager.f, indx=1.5),
                                idx=0, t=imager.sp, insert=True)

    opt_model.ele_model.add_dummy_interface_at_image(seq_model,
                                                     seq_model.gbl_tfrms)

    opt_model.update_model()

    return opt_model


def update_specsheet(iid, opt_model):
    specsheet = opt_model.specsheet
    specsheet_from_parax_data(opt_model, specsheet)
    iid.specsheet_dict[specsheet.conjugate_type] = specsheet
    iid.update_values()


def create_new_ideal_imager(**inputs):
    conj_type = (inputs['conjugate_type'] if 'conjugate_type' in inputs
                 else 'finite')
    specsheets = create_specsheets()
    if 'opt_model' in inputs:
        opt_model = inputs['opt_model']
        specsheet = create_specsheet_from_model(opt_model,
                                                specsheets=specsheets)
        conj_type = specsheet.conjugate_type
        specsheets[conj_type] = specsheet

    if 'gui_parent' in inputs:
        gui_parent = inputs['gui_parent']
        opt_model = gui_parent.app_manager.model
        iid = IdealImagerDialog(conj_type, specsheets,
                                cmd_fct=gui_parent.handle_ideal_imager_command,
                                parent=gui_parent)

        gui_parent.add_subwindow(iid, ModelInfo(gui_parent.app_manager.model,
                                                update_specsheet,
                                                (iid, opt_model)))
        iid.update_values()
        iid.show()
    else:
        iid = IdealImagerDialog(conj_type, specsheets)
        iid.update_values()
        iid.exec_()


def create_yybar_model():
    opt_model = OpticalModel()

    # put in minimum calculation defaults
    opt_model.seq_model.gaps[0].thi = 1.0
    opt_model.optical_spec.field_of_view.type = 'OBJ_HT'
    opt_model.optical_spec.field_of_view.set_from_list([0., 1.])

    opt_model.update_model()

    return opt_model

def get_defaults_from_gui_parent(gui_parent):
    if gui_parent:
        refresh_gui = gui_parent.refresh_gui
        is_dark = gui_parent.is_dark
    else:
        refresh_gui = None
        is_dark = True
    return refresh_gui, is_dark


def create_lens_layout_view(opt_model, gui_parent=None):
    fig = LensLayoutFigure(opt_model)
    view_width = 660
    view_ht = 440
    title = "Lens Layout View"
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht)


def create_live_layout_view(opt_model, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    fig = InteractiveLayout(opt_model, refresh_gui,
                            do_draw_frame=True, do_draw_axes=True,
                            is_dark=is_dark)
    cmds = create_live_layout_commands(fig)
    view_width = 880
    view_ht = 660
    title = "Optical Layout"
    panel_fcts = [create_2d_figure_toolbar,
                  create_draw_rays_groupbox,
                  ]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts,
                              commands=cmds)


def create_live_layout_commands(fig):
    lo = fig.layout
    cmds = []
    # Add thin lens
    cmds.append(('Add Thin Lens', (lo.register_commands, (),
                 {'apply_fct': layout.add_thinlens})))
    # Add lens
    cmds.append(('Add Lens', (lo.register_commands, (),
                 {'apply_fct': layout.add_lens})))
    # Add doublet
    cmds.append(('Add Cemented Doublet', (lo.register_commands, (),
                 {'apply_fct': layout.add_doublet})))
    # Add mirror
    cmds.append(('Add Mirror', (lo.register_commands, (),
                 {'apply_fct': layout.add_mirror,
                  'profile': Spherical})))
    cmds.append(('Add Conic Mirror', (lo.register_commands, (),
                 {'apply_fct': layout.add_conic,
                  'profile': Conic})))
    cmds.append(('Add Parabola', (lo.register_commands, (),
                 {'apply_fct': layout.add_conic,
                  'profile': Conic,
                  'cc': -1.0})))

    return cmds


def create_paraxial_design_view(opt_model, dgm_type, gui_parent=None):
    fig = ParaxialDesignFigure(opt_model, gui_parent.refresh_gui, dgm_type,
                               figsize=(5, 4))
    cmds = diagram.create_parax_design_commands(fig)
    view_width = 650
    view_ht = 500
    title = "Paraxial Design View"
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              commands=cmds)


def create_paraxial_design_view_v2(opt_model, dgm_type, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    fig = dgm.InteractiveDiagram(opt_model, refresh_gui, dgm_type,
                                 do_draw_frame=True, do_draw_axes=True,
                                 aspect='auto', is_dark=is_dark)
    panel_fcts = [create_2d_figure_toolbar,
                  ]
    if dgm_type == 'ht':
        cmds = diagram.create_parax_design_commands(fig)
        panel_fcts.append(create_diagram_controls_groupbox)
    else:
        cmds = None
    view_width = 880
    view_ht = 660
    title = "Paraxial Design View"
    # panel_fcts = [create_2d_figure_toolbar,
    #               create_diagram_controls_groupbox,
    #               ]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts,
                              commands=cmds)


def create_ray_fan_view(opt_model, data_type, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    fig = RayFanFigure(opt_model, data_type,
                       scale_type=Fit.All_Same,
                       figsize=(5, 4), dpi=100, is_dark=is_dark)
    view_width = 600
    view_ht = 600
    if data_type == "Ray":
        title = "Ray Fan View"
    elif data_type == "OPD":
        title = "OPD Fan View"
    else:
        title = "bad data_type argument"
    panel_fcts = [create_plot_scale_panel]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts)


def create_ray_grid_view(opt_model, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    num_flds = len(opt_model.optical_spec.field_of_view.fields)

    fig = SpotDiagramFigure(opt_model, scale_type=Fit.All_Same,
                            dpi=100, is_dark=is_dark)
    view_box = 300
    view_width = view_box
    view_ht = num_flds * view_box
    title = "Spot Diagram"
    panel_fcts = [create_plot_scale_panel]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts)


def create_wavefront_view(opt_model, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    num_flds = len(opt_model.optical_spec.field_of_view.fields)
    num_wvls = len(opt_model.optical_spec.spectral_region.wavelengths)

    fig = WavefrontFigure(opt_model, scale_type=Fit.All_Same,
                          num_rays=32, dpi=100, is_dark=is_dark)
#                                 figsize=(2*num_wvls, 2*num_flds))
    view_box = 300
    view_width = num_wvls * view_box
    view_ht = num_flds * view_box
    title = "Wavefront Map"
    panel_fcts = [create_plot_scale_panel]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts)


def create_field_curves(opt_model, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    fig = FieldCurveFigure(opt_model, dpi=100, is_dark=is_dark)
    view_width = 600
    view_ht = 600
    title = "Field Curves"
    panel_fcts = [create_plot_scale_panel]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts)


def create_3rd_order_bar_chart(opt_model, gui_parent=None):
    refresh_gui, is_dark = get_defaults_from_gui_parent(gui_parent)
    fig = ThirdOrderBarChart(opt_model, dpi=100, is_dark=is_dark)
    view_width = 600
    view_ht = 600
    title = "3rd Order Aberrations"
    panel_fcts = [create_plot_scale_panel]
    plotview.create_plot_view(gui_parent, fig, title, view_width, view_ht,
                              add_panel_fcts=panel_fcts)


def update_table_view(table_view):
    table_model = table_view.model()
    table_model.endResetModel()


def create_lens_table_model(seq_model):
    colEvalStr = ['.ifcs[{}].interface_type()',
                  '.ifcs[{}].profile_cv',
                  '.ifcs[{}].surface_od()', '.gaps[{}].thi',
                  '.gaps[{}].medium.name()', '.ifcs[{}].interact_mode']
    rowHeaders = seq_model.surface_label_list()
    colHeaders = ['type', 'cv', 'sd', 'thi', 'medium', 'mode']
    colFormats = ['{:s}', '{:12.7g}', '{:12.5g}', '{:12.5g}',
                  '{:s}', '{:s}']
    return PyTableModel(seq_model, '', colEvalStr, rowHeaders,
                        colHeaders, colFormats, True,
                        get_num_rows=seq_model.get_num_surfaces,
                        get_row_headers=seq_model.surface_label_list)


def create_element_table_model(opt_model):
    ele_model = opt_model.ele_model

    def get_row_headers():
        return [str(i) for i in range(ele_model.get_num_elements())]
    colEvalStr = ['.elements[{}].label', '.element_type({})',
                  '.elements[{}].medium_name',
                  '.elements[{}].tfrm[1][1]',
                  '.elements[{}].tfrm[1][2]',
                  '.elements[{}].reference_idx()',
                  '.elements[{}].reference_interface().interface_type()']

    rowHeaders = get_row_headers()
    colHeaders = ['label', 'type', 'medium', 'y', 'z', 'idx', 'ref ifc']
    colFormats = ['{:s}', '{:s}', '{:s}', '{:12.5g}', '{:12.5g}',
                  '{:d}', '{:s}']
    return PyTableModel(ele_model, '', colEvalStr, rowHeaders,
                        colHeaders, colFormats, True,
                        get_num_rows=ele_model.get_num_elements,
                        get_row_headers=get_row_headers)


def create_ray_table_model(opt_model, ray):
    colEvalStr = ['[{}].p[0]', '[{}].p[1]', '[{}].p[2]',
                  '[{}].d[0]', '[{}].d[1]', '[{}].d[2]',
                  '[{}].dst']
    seq_model = opt_model.seq_model
    rowHeaders = seq_model.surface_label_list()
    colHeaders = ['x', 'y', 'z', 'l', 'm', 'n', 'length']
    colFormats = ['{:12.5g}', '{:12.5g}', '{:12.5g}', '{:9.6f}',
                  '{:9.6f}', '{:9.6f}', '{:12.5g}']
    return PyTableModel(ray, '', colEvalStr, rowHeaders,
                        colHeaders, colFormats, False,
                        get_num_rows=seq_model.get_num_surfaces,
                        get_row_headers=seq_model.surface_label_list)


def create_parax_table_model(opt_model):
    rootEvalStr = ".optical_spec.parax_data"
    colEvalStr = ['[0][{}][0]', '[0][{}][1]', '[0][{}][2]',
                  '[1][{}][0]', '[1][{}][1]', '[1][{}][2]']
    seq_model = opt_model.seq_model
    rowHeaders = seq_model.surface_label_list()
    colHeaders = ['y', 'u', 'i', 'y-bar', 'u-bar', 'i-bar']
    colFormats = ['{:12.5g}', '{:9.6f}', '{:9.6f}', '{:12.5g}',
                  '{:9.6f}', '{:9.6f}']
    return PyTableModel(opt_model, rootEvalStr, colEvalStr, rowHeaders,
                        colHeaders, colFormats, False,
                        get_num_rows=seq_model.get_num_surfaces,
                        get_row_headers=seq_model.surface_label_list)


def create_parax_model_table(opt_model):
    rootEvalStr = ".parax_model"
    colEvalStr = ['.ax[{}][0]', '.pr[{}][0]', '.ax[{}][1]', '.pr[{}][1]',
                  '.sys[{}][0]', '.sys[{}][1]', '.sys[{}][2]', '.sys[{}][3]']
    seq_model = opt_model.seq_model
    rowHeaders = seq_model.surface_label_list()
    colHeaders = ['y', 'y-bar', 'u', 'u-bar', 'pwr', 'tau', 'n after', 'mode']
    colFormats = ['{:12.5g}', '{:12.5g}', '{:9.6f}', '{:9.6f}',
                  '{:12.7g}', '{:12.5g}', '{:7.4f}', '{:s}']
    return PyTableModel(opt_model, rootEvalStr, colEvalStr, rowHeaders,
                        colHeaders, colFormats, True,
                        get_num_rows=seq_model.get_num_surfaces,
                        get_row_headers=seq_model.surface_label_list)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Module for thin lens interface type

.. Created on Wed May 16 14:05:38 2018

.. codeauthor: Michael J. Hayford
"""


import numpy as np
from rayoptics.optical.surface import Interface
from rayoptics.optical.doe import HolographicElement


class ThinLens(Interface):
    def __init__(self, lbl='', power=0.0, ref_index=1.5, **kwargs):
        super().__init__(interact_mode='transmit',
                         phase_element=HolographicElement(),
                         **kwargs)
        self.label = lbl
        self.optical_power = power
        self.ref_index = ref_index
        self.bending = 0.0

    def list_thinlens(self):
        if len(self.label) > 0:
            print(self.label)
        print("power: {:12.6g}".format(self.optical_power))
        self.phase_element.list_hoe()

    def __repr__(self):
        if len(self.label) > 0:
            return "ThinLens(%r: power=%r)" % (self.label, self.optical_power)
        else:
            return "ThinLens(power=%r)" % (self.optical_power)

    def update(self):
        super().update()

    def full_profile(self, sd, flat_id=None, dir=1, steps=6):
        prf = []
        if len(sd) == 1:
            sd_lwr = -sd[0]
            sd_upr = sd[0]
        else:
            sd_lwr = sd[0]
            sd_upr = sd[1]

        prf.append([0, dir*sd_lwr])
        prf.append([0, dir*sd_upr])

        return prf

    @property
    def profile_cv(self):
        return self._power

    @profile_cv.setter
    def profile_cv(self, cv):
        self._power = cv

    def surface_od(self):
        return self.max_aperture

    def set_max_aperture(self, max_ap):
        super().set_max_aperture(max_ap)

    @property
    def optical_power(self):
        return self._power

    @optical_power.setter
    def optical_power(self, pwr):
#        print("optical_power {}: pwr={}, {} obj={}, {}".format(self.label,
#              self._power, pwr, self.phase_element.obj_pt[2], 1./pwr))
        self._power = pwr
        try:
            self.phase_element.obj_pt[2] = 1./pwr
        except ZeroDivisionError:
            self.phase_element.obj_pt[2] = 1e+10
        finally:
            self.phase_element.obj_virtual = True if pwr > 0. else False

    def set_optical_power(self, pwr, n_before, n_after):
        self.delta_n = n_after - n_before
        self.optical_power = pwr

    def apply_scale_factor(self, scale_factor):
        super().apply_scale_factor(scale_factor)
        self.optical_power = self.optical_power/scale_factor

    def from_first_order(self, nu_before, nu_after, y):
        # nu_before used for reference point
        ref = -y/nu_before if nu_before != 0.0 else 1e+10
        obj = -y/nu_after if nu_after != 0.0 else 1e+10
#        pm = self.phase_element
#        print("from_first_order {}:\n pwr={}, {}\n ref={} ({}), {} ({})"
#              "\n obj={} ({}), {} ({})"
#              .format(self.label,
#                      self._power, (nu_before - nu_after)/y,
#                      pm.ref_pt[2], pm.ref_virtual, ref, ref > 0.,
#                      pm.obj_pt[2], pm.obj_virtual, obj, obj > 0.))
        self.phase_element.ref_pt[2] = ref
        self.phase_element.ref_virtual = True if ref > 0. else False
        self.phase_element.obj_pt[2] = obj
        self.phase_element.obj_virtual = True if obj > 0. else False
        self._power = (nu_before - nu_after)/y

    def normal(self, p):
        return np.array([0., 0., 1.])

    def intersect(self, p0, d, **kwargs):
        s1 = -p0[2]/d[2]
        p = p0 + s1*d
        return s1, p

    def phase(self, pt, d_in, normal, wl):
        return self.phase_element.phase(pt, d_in, normal)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" Qt5 dialog box for ideal imager ui

.. Created on Tue Jun 18 13:09:19 2019

.. codeauthor: Michael J. Hayford
"""

from rayoptics.optical.specsheet import create_specsheets
from rayoptics.optical import idealimager
from rayoptics.optical.etendue import (obj_img_set, fld_ape_set,
                                       fld_labels, ap_labels)

from rayoptics.util.dict2d import dict2D

from PyQt5.QtCore import Qt as qt
from PyQt5.QtWidgets import (QApplication, QDialog, QRadioButton, QWidget,
                             QStackedWidget, QGridLayout, QGroupBox,
                             QHBoxLayout, QLabel, QLineEdit, QCheckBox,
                             QVBoxLayout, QDialogButtonBox)


def value_to_text(value, fmt_str="{:> #.5f}"):
    if value is None:
        value_text = ''
    elif type(value) is str:
        value_text = value
    else:
        value_text = fmt_str.format(value)
    return value_text


class IdealImagerDialog(QWidget):
    def __init__(self, conjugate_type, specsheets, cmd_fct=None,
                 **kwargs):
        super().__init__(**kwargs)

        self.dlog_attrs = {}

        self.conjugate_type = conjugate_type
        self.conjugate_box = self.createConjugateBox(itype=self.conjugate_type)

        self.specsheet_dict = specsheets

        self.imager_stack = {}
        self.etendue_stack = {}
        self.imager_groupbox_stack = QStackedWidget()
        self.etendue_groupbox_stack = QStackedWidget()

        for key, specsheet in specsheets.items():
            isgb = ImagerSpecGroupBox(self, specsheet)
            self.imager_stack[key] = isgb
            self.imager_groupbox_stack.addWidget(isgb)
            egb = EtendueGroupBox(self, key, specsheet)
            self.etendue_stack[key] = egb
            self.etendue_groupbox_stack.addWidget(egb)

        imager_groupbox = self.imager_stack[self.conjugate_type]
        self.imager_groupbox_stack.setCurrentWidget(imager_groupbox)

        etendue_groupbox = self.etendue_stack[self.conjugate_type]
        self.etendue_groupbox_stack.setCurrentWidget(etendue_groupbox)

        # construct the top level layout
        overallLayout = QVBoxLayout(self)

        mainLayout = QHBoxLayout()
        mainLayout.addWidget(self.conjugate_box)
        mainLayout.addWidget(self.imager_groupbox_stack)
        mainLayout.addWidget(self.etendue_groupbox_stack)

        self.button_box = self.createButtonBox(cmd_fct)
        overallLayout.addLayout(mainLayout)
        overallLayout.addWidget(self.button_box)

        self.setLayout(overallLayout)

        self.setWindowTitle("Optical Spec Sheet")

    def createButtonBox(self, cmd_fct):
        def clicked(button):
            command = button.text()
            specsheet = self.specsheet_dict[self.conjugate_type]
            if cmd_fct:
                cmd_fct(self, command, specsheet)
            else:
                print(button.text(), 'button pressed')

        buttonbox = QDialogButtonBox(qt.Horizontal, self)
        buttonbox.addButton('New', QDialogButtonBox.ApplyRole)
        buttonbox.addButton(QDialogButtonBox.Apply)
        buttonbox.addButton('Update', QDialogButtonBox.ApplyRole)
        buttonbox.addButton(QDialogButtonBox.Close)
        for b in buttonbox.buttons():
            b.setAutoDefault(False)
#        buttonbox.setCenterButtons(True)
        buttonbox.clicked.connect(clicked)
        return buttonbox

    def createConjugateBox(self, itype='infinite'):
        conjugate_box = QGroupBox("Conjugates")
        layout = QVBoxLayout()

        infinite = QRadioButton('Infinite')
        self.dlog_attrs['infinite'] = infinite
        layout.addWidget(infinite)
        infinite.clicked.connect(lambda: self.change_conjugate('infinite'))

        finite = QRadioButton('Finite')
        self.dlog_attrs['finite'] = finite
        layout.addWidget(finite)
        finite.clicked.connect(lambda: self.change_conjugate('finite'))

        conjugate_box.setLayout(layout)

        self.dlog_attrs[itype].setChecked(True)

        return conjugate_box

    def change_conjugate(self, conj_type):
        print("change_conjugate:", conj_type)
        if self.conjugate_type is not conj_type:
            self.update_conjugate(conj_type)

    def update_conjugate(self, conj_type):
        prev_specsheet = self.specsheet_dict[self.conjugate_type]
        self.conjugate_type = conj_type

        new_specsheet = self.specsheet_dict[conj_type]
        new_partition, max_inputs = new_specsheet.partition_defined()
        if new_partition is None:
            if prev_specsheet.imager.f is not None:
                prev_f = prev_specsheet.imager.f
                new_specsheet.imager_inputs['f'] = prev_f

        new_imager_groupbox = self.imager_stack[conj_type]
        self.imager_groupbox_stack.setCurrentWidget(new_imager_groupbox)

        etendue_groupbox = self.etendue_stack[conj_type]
        self.etendue_groupbox_stack.setCurrentWidget(etendue_groupbox)

        self.update_values()

    def update_values(self):
        """ callback routine for any dialog value change """
        conj_type = self.conjugate_type
        specsheet = self.specsheet_dict[conj_type]

        imager_groupbox = self.imager_stack[conj_type]
        etendue_groupbox = self.etendue_stack[conj_type]

        imager_inputs = imager_groupbox.specsheet.imager_inputs
        etendue_inputs = etendue_groupbox.specsheet.etendue_inputs
        imager, etendue_values = specsheet.generate_from_inputs(imager_inputs,
                                                                etendue_inputs)

        specsheet.imager_inputs = imager_inputs
        specsheet.etendue_inputs = etendue_inputs

        imager_groupbox.update_values()
        etendue_groupbox.update_values()

    def update_checkboxes(self):
        """ callback routine for any dialog checkbox change """
        imager_groupbox = self.imager_stack[self.conjugate_type]
        etendue_groupbox = self.etendue_stack[self.conjugate_type]

        imager_groupbox.update_checkboxes()
        etendue_groupbox.update_checkboxes()


class ImagerSpecGroupBox(QGroupBox):
    def __init__(self, parent, specsheet, keys=None, labels=None, **kwargs):
        super().__init__(title='Imager specs', **kwargs)

        self.parent = parent
        self.specsheet = specsheet

        self.dlog_attrs = {}

        self.keys = keys if keys else idealimager.ideal_imager_keys
        self.labels = labels if labels else idealimager.ideal_imager_labels

        layout = QGridLayout()
        for i, key in enumerate(self.keys):
            label = QLabel(self.labels[i]+':')
            lineEdit = QLineEdit()
            checkBox = QCheckBox()
            row = i + 1
            layout.addWidget(label, row, 0)
            layout.addWidget(lineEdit, row, 1)
            layout.addWidget(checkBox, row, 2)
            self.dlog_attrs[key] = (label, lineEdit, checkBox)
            lineEdit.setText(value_to_text(specsheet.imager[i]))
            # Use returnPressed signal rather than editingFinished. The latter
            #  fires on change of focus as well as return. We only want a
            #  signal when the lineEdit contents change and are committed by
            #  the user.
            lineEdit.returnPressed.connect(lambda k=key: self.value_change(k))

            if key in specsheet.imager_inputs:
                if not checkBox.isChecked():
                    checkBox.setCheckState(qt.Checked)
            checkBox.stateChanged.connect(lambda state, k=key:
                                          self.chkbox_change(state, k))

        self.setLayout(layout)
        self.update_values()

    def value_change(self, imager_key):
        label, lineEdit, checkBox = self.dlog_attrs[imager_key]

        try:
            value = float(lineEdit.text())
        except ValueError:
            value = None
        finally:
            imager_inputs = self.specsheet.imager_inputs
            if value is not None:
                if imager_key in imager_inputs or len(imager_inputs) < 2:
                    imager_inputs[imager_key] = value
                    if not checkBox.isChecked():
                        checkBox.setCheckState(qt.Checked)
            else:
                if imager_key in imager_inputs:
                    del imager_inputs[imager_key]
                    if checkBox.isChecked():
                        checkBox.setChecked(False)
            self.parent.update_values()

    def chkbox_change(self, state, imager_key):
        label, lineEdit, checkBox = self.dlog_attrs[imager_key]

        checked = state == qt.Checked
        if checked:
            try:
                value = float(lineEdit.text())
            except ValueError:
                value = None
            finally:
                self.specsheet.imager_inputs[imager_key] = value
                if not checkBox.isChecked():
                    checkBox.setChecked(True)
        else:
            if imager_key in self.specsheet.imager_inputs:
                del self.specsheet.imager_inputs[imager_key]
                if checkBox.isChecked():
                    checkBox.setChecked(False)
        self.parent.update_checkboxes()

    def update_values(self):
        if self.specsheet.imager is not None:
            value_list = self.specsheet.imager
            # refill gui widgets and set enabled state
            for i, key in enumerate(self.keys):
                value = value_list[i]
                label, lineEdit, checkBox = self.dlog_attrs[key]
                lineEdit.setText(value_to_text(value))

        self.update_checkboxes()

    def update_checkboxes(self):
        partition, max_inputs = self.specsheet.partition_defined()
        if partition is 'imager':
            # the inputs for imager are sufficient to calculate all; mark
            # inputs editable and disable the others
            enabled_list = [False]*5
            for key in self.specsheet.imager_inputs:
                enabled_list[self.keys.index(key)] = True
        elif bool(partition):
            # an imager is defined by other partitions, this means:
            #  1) the imager key is not editable
            #  2a) if one imager input, the remainder are uneditable
            #  2b) if no imager inputs, all but the imager key is editable
            if len(self.specsheet.imager_inputs):
                enabled_list = [False]*5
                for key in self.specsheet.imager_inputs:
                    enabled_list[self.keys.index(key)] = True
                imager_key = self.specsheet.imager_defined()
                if imager_key:
                    enabled_list[self.keys.index(imager_key)] = False
            else:
                enabled_list = [True]*5
                imager_key = self.specsheet.imager_defined()
                if imager_key:
                    enabled_list[self.keys.index(imager_key)] = False
        else:
            # no imager defined, all items are editable
            enabled_list = [True]*5

        # update the ui from the enabled list and inputs
        for i, key in enumerate(self.keys):
            label, lineEdit, checkBox = self.dlog_attrs[key]

            if key in self.specsheet.imager_inputs:
                if not checkBox.isChecked():
                    checkBox.setCheckState(qt.Checked)
            else:
                if checkBox.isChecked():
                    checkBox.setCheckState(qt.Unchecked)

            if self.specsheet.frozen_imager_inputs[i]:
                enabled_list[i] = False

            lineEdit.setEnabled(enabled_list[i])
            checkBox.setEnabled(enabled_list[i])


class EtendueGroupBox(QGroupBox):
    def __init__(self, parent, itype, specsheet, **kwargs):
        super().__init__(title='etendue definition', **kwargs)

        self.parent = parent
        self.specsheet = specsheet

        self.groupboxes = dict2D(fld_ape_set, obj_img_set)

        if itype is 'infinite':
            gb00 = SpaceGroupBox(parent=self, title='object space',
                                 obj_img='object', fld_ape='field',
                                 keys=(fld_labels[1],))
            self.groupboxes['field']['object'] = gb00
            gb10 = SpaceGroupBox(parent=self, title='',
                                 obj_img='object', fld_ape='aperture',
                                 keys=(ap_labels[0],))
            self.groupboxes['aperture']['object'] = gb10
            gb01 = SpaceGroupBox(parent=self, title='image space',
                                 obj_img='image', fld_ape='field',
                                 keys=(fld_labels[0],))
            self.groupboxes['field']['image'] = gb01
            gb11 = SpaceGroupBox(parent=self, title='',
                                 obj_img='image', fld_ape='aperture',
                                 keys=(ap_labels[2], ap_labels[1]))
            self.groupboxes['aperture']['image'] = gb11

        else:
            gb00 = SpaceGroupBox(parent=self, title='object space',
                                 fld_ape='field', obj_img='object',
                                 keys=(fld_labels[0],))
            self.groupboxes['field']['object'] = gb00
            gb10 = SpaceGroupBox(parent=self, title='',
                                 fld_ape='aperture', obj_img='object',
                                 keys=(ap_labels[2], ap_labels[1]))
            self.groupboxes['aperture']['object'] = gb10
            gb01 = SpaceGroupBox(parent=self, title='image space',
                                 fld_ape='field', obj_img='image',
                                 keys=(fld_labels[0],))
            self.groupboxes['field']['image'] = gb01
            gb11 = SpaceGroupBox(parent=self, title='',
                                 fld_ape='aperture', obj_img='image',
                                 keys=(ap_labels[2], ap_labels[1]))
            self.groupboxes['aperture']['image'] = gb11

        layout = QGridLayout()
        layout.addWidget(gb00, 0, 0)
        layout.addWidget(gb10, 1, 0)
        layout.addWidget(gb01, 0, 1)
        layout.addWidget(gb11, 1, 1)

        self.setLayout(layout)

    def value_change(self, fld_ape, obj_img, key):
        """ callback routine for item value widget """
        inputs = self.specsheet.etendue_inputs[fld_ape][obj_img]
        gb = self.groupboxes[fld_ape][obj_img]
        label, lineEdit, checkBox = gb.dlog_attrs[key]
        try:
            value = float(lineEdit.text())
        except ValueError:
            return
        else:
            if (key in inputs or len(inputs) < 2):
                inputs[key] = value
                if not checkBox.isChecked():
                    checkBox.setCheckState(qt.Checked)
            self.parent.update_values()

    def chkbox_change(self, state, fld_ape, obj_img, key):
        """ callback routine for item checkbox """
        inputs = self.specsheet.etendue_inputs[fld_ape][obj_img]
        gb = self.groupboxes[fld_ape][obj_img]
        label, lineEdit, checkBox = gb.dlog_attrs[key]
        checked = state == qt.Checked
        if checked:
            try:
                value = float(lineEdit.text())
            except ValueError:
                value = None
            finally:
                inputs[key] = value
                if not checkBox.isChecked():
                    checkBox.setChecked(True)
        else:
            if key in inputs:
                del inputs[key]
                if checkBox.isChecked():
                    checkBox.setChecked(False)
        self.parent.update_checkboxes()

    def update_values(self):
        etendue_inputs = self.specsheet.etendue_inputs
        etendue_values = self.specsheet.etendue_values

        for fld_ape_key, fld_ape_value in etendue_values.items():
            for obj_img_key, values in fld_ape_value.items():
                sgb = self.groupboxes[fld_ape_key][obj_img_key]
                inputs = etendue_inputs[fld_ape_key][obj_img_key]
                sgb.update_values(inputs, values)
        self.update_checkboxes()

    def update_checkboxes(self):
        """ update the enabled and checked state for the etendue groupbox """
        etendue_inputs = self.specsheet.etendue_inputs
        etendue_values = self.specsheet.etendue_values
        partition, max_inputs = self.specsheet.partition_defined()

        for fld_ape_key, fld_ape_value in etendue_values.items():
            num_fld_ape_inputs = self.specsheet.partitions[fld_ape_key]
            for obj_img_key, values in fld_ape_value.items():
                sgb = self.groupboxes[fld_ape_key][obj_img_key]
                inputs = etendue_inputs[fld_ape_key][obj_img_key]
                if num_fld_ape_inputs == 2:
                    partition_defined = True
                elif num_fld_ape_inputs == 1 and len(inputs) == 1:
                    partition_defined = True
                elif num_fld_ape_inputs == 1 and len(inputs) == 0:
                    partition_defined = True if bool(partition) else False
                else:  # num_fld_ape_inputs == 0 and len(inputs) == 0
                    partition_defined = False

                sgb.update_checkboxes(inputs, values,
                                      partition_defined=partition_defined)


class SpaceGroupBox(QGroupBox):
    def __init__(self, parent, title, fld_ape, obj_img, keys,
                 labels=None, **kwargs):
        super().__init__(title=title, **kwargs)
        self.dlog_attrs = {}

        self.parent = parent
        self.fld_ape = fld_ape
        self.obj_img = obj_img

        self.keys = keys
        if labels is None:
            self.labels = keys
        else:
            self.labels = labels

        layout = QGridLayout()

        for i, key in enumerate(self.keys):
            label = QLabel(self.labels[i]+':')
            lineEdit = QLineEdit()
            checkBox = QCheckBox()
            row = i + 1
            layout.addWidget(label, row, 0)
            layout.addWidget(lineEdit, row, 1)
            layout.addWidget(checkBox, row, 2)
            self.dlog_attrs[key] = (label, lineEdit, checkBox)

            # Use returnPressed signal rather than editingFinished. The latter
            #  fires on change of focus as well as return. We only want a
            #  signal when the lineEdit contents change and are committed by
            #  the user.
            lineEdit.returnPressed.connect(lambda k=key:
                                           parent.value_change(fld_ape,
                                                               obj_img, k))

            checkBox.stateChanged.connect(lambda state, k=key:
                                          parent.chkbox_change(state, fld_ape,
                                                               obj_img, k))

        self.setLayout(layout)

    def update_values(self, inputs, values):
        """ update the display for the etendue cell being updated """
        for key in self.keys:
            value = values[key]
            label, lineEdit, checkBox = self.dlog_attrs[key]
            lineEdit.setText(value_to_text(value))

    def update_checkboxes(self, inputs, values, partition_defined=False):
        """ update the display for the etendue cell being updated

        A partition is an aperture or field pair of object/image inputs.

        If it is defined, this means that all attrs can be supplied.
        In this case, the inputs will have editable values and (checked)
        checkboxes; the remaining attrs will have uneditable values and
        (unchecked) checkboxes.

        If the partition is not defined, all values and checkboxes will be
        editable, the input attrs, if any, will be checked.
        """
        if partition_defined:
            for key in self.keys:
                label, lineEdit, checkBox = self.dlog_attrs[key]

                if key in inputs:
                    lineEdit.setEnabled(True)
                    checkBox.setEnabled(True)
                    if not checkBox.isChecked():
                        checkBox.setCheckState(qt.Checked)
                else:
                    lineEdit.setEnabled(False)
                    checkBox.setEnabled(False)
                    if checkBox.isChecked():
                        checkBox.setCheckState(qt.Unchecked)

        else:
            for key in self.keys:
                label, lineEdit, checkBox = self.dlog_attrs[key]
                lineEdit.setEnabled(True)
                checkBox.setEnabled(True)

                if key in inputs:
                    if not checkBox.isChecked():
                        checkBox.setCheckState(qt.Checked)
                else:
                    if checkBox.isChecked():
                        checkBox.setCheckState(qt.Unchecked)


if __name__ == '__main__':

    import sys

    class Dialog2(QDialog):
        def __init__(self, parent, conjugate_type, specsheets):
            QDialog.__init__(self, parent)
#            self.setModal(0)
            iid = IdealImagerDialog(conjugate_type, specsheets, parent=self)
            iid.update_values()

        def exit(self):
            self.close()

    app = QApplication(sys.argv)
    specsheets = create_specsheets()
    dialog = Dialog2(None, 'infinite', specsheets)
    dialog.show()
#    dialog.exec()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Qt GraphicsItem subclasses for 2D graphics views

.. Created on Tue Mar 20 10:39:27 2018

.. codeauthor: Michael J. Hayford
"""

from PyQt5.QtCore import QPointF
from PyQt5.QtGui import (QColor, QPen, QPolygonF)
from PyQt5.QtWidgets import QGraphicsPolygonItem

import rayoptics.gui.layout as layout


class OpticalElement(QGraphicsPolygonItem):
    """ QGraphicsItem subclass for 2D lens element graphics """
    def __init__(self, element):
        super().__init__()

        self.oe = layout.OpticalElement(element)
        self.update_shape()

        color = self.oe.render_color()
        self.setBrush(QColor(*color))
        pen = QPen()
        pen.setCosmetic(True)
        self.setPen(pen)

    def update(self):
        super().update()
        print("in update")

    def update_shape(self):
        poly, bbox = self.oe.update_shape()
        polygon = QPolygonF()
        for p in poly:
            # Qt canvas coordinates have 0,0 at top left of window, i.e. +y is
            #  pointing down
            polygon.append(QPointF(p[0], -p[1]))

        self.setPolygon(polygon)


class RayBundle(QGraphicsPolygonItem):
    """ QGraphicsItem subclass for ray bundle from a single field point """
    def __init__(self, opt_model, field_num, start_offset):
        super().__init__()
        fld, wvl, foc = opt_model.optical_spec.lookup_fld_wvl_focus(field_num)
        self.rb = layout.RayBundle(opt_model, fld, wvl, start_offset)
        self.update_shape()

        self.setBrush(QColor(254, 197, 254, 64))  # magenta, 25%
        pen = QPen()
        pen.setCosmetic(True)
        self.setPen(pen)

    def update_shape(self):
        poly, bbox = self.rb.update_shape()
        polygon = QPolygonF()
        for p in poly:
            # Qt canvas coordinates have 0,0 at top left of window, i.e. +y is
            #  pointing down
            polygon.append(QPointF(p[0], -p[1]))

        self.setPolygon(polygon)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 18 10:59:29 2017

@author: Mike
"""


import unittest
from pytest import approx
from rayoptics.optical.profiles import Spherical, EvenPolynomial
from rayoptics.util.misc_math import normalize
import numpy as np
import numpy.testing as npt
from math import sqrt


class SphericalProfileTestCase(unittest.TestCase):
    def setUp(self):
        self.dir0 = np.array([0., 0., 1.])
        self.p0 = np.array([0., 0., -1.])
        self.p1 = np.array([0., 1., -1.])
        self.eps = 1.0e-12
        self.z_dir = 1.0

    def test_normalize(self):
        v0 = normalize(np.array([0., 0., 0.]))
        npt.assert_array_equal(v0, np.array([0., 0., 0.]))

        v1 = normalize(np.array([1., 1., 1.]))
        sqrRt3 = sqrt(3)/3
        npt.assert_allclose(v1, np.array([sqrRt3, sqrRt3, sqrRt3]),
                            rtol=1e-14)

    def test_planar_sphere(self):
        s1 = Spherical(c=0.0)
        p_truth = 1.0, np.array([0., 0., 0.])

        # test for p0 ray, (0, 0, 0), (0, 0, 1)
        p0s1 = s1.intersect(self.p0, self.dir0, self.eps, self.z_dir)
        assert (p0s1[0], p0s1[1].all()) == (p_truth[0], p_truth[1].all())

        # test for p1 ray, (0, 1, 0), (0, 0, 1)
        p1s1 = s1.intersect(self.p1, self.dir0, self.eps, self.z_dir)
        assert ((p1s1[0], p1s1[1].all()) ==
                (p_truth[0], np.array([0., 1., 0.]).all()))

    def test_convex_sphere(self):
        r2 = 10
        c2 = 1/r2
        s2 = Spherical(c=c2)

        # test for p0 ray, (0, 0, 0), (0, 0, 1)
        p0_truth = 1.0, np.array([0., 0., 0.])

        p0s2_dir = s2.intersect(self.p0, self.dir0, self.eps, self.z_dir)
        p0s2_tnp = s2.intersect_tangent_plane(self.p0, self.dir0,
                                              self.eps, self.z_dir)
        assert ((p0s2_dir[0], p0s2_dir[1].all()) ==
                (p0s2_tnp[0], p0s2_tnp[1].all()))

        p0s2 = p0s2_dir
        assert ((p0s2[0], p0s2[1].all()) == (p0_truth[0], p0_truth[1].all()))

        # A spherical EvenPolynomial will use iteration to find the
        #  intersection. Use this case to further check results
        sa2 = EvenPolynomial(c=c2)
        p0sa2 = sa2.intersect(self.p0, self.dir0, self.eps, self.z_dir)
        assert ((p0sa2[0], p0sa2[1].all()) ==
                (p0_truth[0], p0_truth[1].all()))

        assert (p0s2_dir[0], p0s2_dir[1].all()) == (p0sa2[0], p0sa2[1].all())

        # test for p1 ray, (0, 1, 0), (0, 0, 1)
        sag2 = r2 - sqrt(r2*r2 - 1.0)
        p1_truth = 1+sag2, np.array([0., 1., sag2])

        p1s2_dir = s2.intersect(self.p1, self.dir0, self.eps, self.z_dir)
        p1s2_tnp = s2.intersect_tangent_plane(self.p1, self.dir0,
                                              self.eps, self.z_dir)
        assert ((p1s2_dir[0], p1s2_dir[1].all()) ==
                (p1s2_tnp[0], p1s2_tnp[1].all()))

        p1s2 = p1s2_dir
        assert ((p1s2[0], p1s2[1].all()) ==
                (p1_truth[0], p1_truth[1].all()))

        dir_p1s2 = s2.normal(p1s2[1])
        dir_p1s2_truth = -normalize(p1_truth[1] - np.array([0, 0, r2]))
        assert dir_p1s2.all() == dir_p1s2_truth.all()

    def test_concave_sphere(self):
        r3 = 10
        s3 = Spherical(r=-r3)

        # test for p0 ray, (0, 0, 0), (0, 0, 1)
        p0_truth = 1.0, np.array([0., 0., 0.])

        p0s3_dir = s3.intersect(self.p0, self.dir0, self.eps, self.z_dir)
        p0s3_tnp = s3.intersect_tangent_plane(self.p0, self.dir0,
                                              self.eps, self.z_dir)
        assert ((p0s3_dir[0], p0s3_dir[1].all()) ==
                (p0s3_tnp[0], p0s3_tnp[1].all()))

        p0s3 = p0s3_dir
        assert ((p0s3[0], p0s3[1].all()) == (p0_truth[0], p0_truth[1].all()))

        # test for p1 ray, (0, 1, 0), (0, 0, 1)
        sag3 = r3 - sqrt(r3*r3 - 1.0)
        p1_truth = 1-sag3, np.array([0., 1., -sag3])

        p1s3_tnp = s3.intersect_tangent_plane(self.p1, self.dir0,
                                              self.eps, self.z_dir)
        p1s3_dir = s3.intersect(self.p1, self.dir0, self.eps, self.z_dir)
        assert ((p1s3_dir[0], p1s3_dir[1].all()) ==
                (p1s3_tnp[0], p1s3_tnp[1].all()))

        p1s3 = p1s3_dir
        assert p1s3[0] == approx(p1_truth[0], rel=1e-14, abs=1e-14)
        assert p1s3[1].all() == p1_truth[1].all()

        dir_p1s3 = s3.normal(p1s3[1])
        dir_p1s3_truth = normalize(p1_truth[1] - np.array([0, 0, -r3]))
        assert dir_p1s3.all() == dir_p1s3_truth.all()

    def test_dbgauss_s1(self):
        r1 = 56.20238
        c1 = 1/r1
        y0 = 25.0
        s1 = Spherical(c=c1)
        p1 = np.array([0., y0, 0.])
        sag1 = r1 - sqrt(r1*r1 - y0*y0)

        # test for p0 ray, (0, 0, 0), (0, 0, 1)
        p0_truth = 1.0, np.array([0., 0., 0.])

        p0s1 = s1.intersect(self.p0, self.dir0, self.eps, self.z_dir)
        assert ((p0s1[0], p0s1[1].all()) ==
                (p0_truth[0], p0_truth[1].all()))
        self.assertEqual(p0s1[0], 1.0)
        npt.assert_allclose(p0s1[1], np.array([0., 0., 0.]),
                            rtol=1e-14, atol=1e-14)

        # test for p1 ray, (0, 1, 0), (0, 0, 1)
        p1_truth = 5.866433424372758, np.array([0., y0, sag1])
        p1s1 = s1.intersect(p1, self.dir0, self.eps, self.z_dir)
        assert p1s1[0] == approx(p1_truth[0], rel=1e-14, abs=1e-14)
        npt.assert_allclose(p1s1[1], p1_truth[1], rtol=1e-14)

        dir_p1s1 = s1.normal(p1s1[1])
        dir_p1s1_truth = -normalize(p1_truth[1] - np.array([0, 0, r1]))
        assert dir_p1s1.all() == dir_p1s1_truth.all()
        npt.assert_allclose(dir_p1s1, dir_p1s1_truth, rtol=1e-14)


if __name__ == '__main__':
    unittest.main(verbosity=3)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Supports model ray tracing in terms of relative aperture and field.

.. Created on Mon Sep 17 23:10:59 2018

.. codeauthor: Michael J. Hayford
"""

import itertools
import logging
import math
import numpy as np
from numpy.linalg import norm
from scipy.optimize import newton, fsolve
from collections import namedtuple
import pandas as pd
import attr

from . import raytrace as rt
from . import model_constants as mc
from .traceerror import TraceError, TraceMissedSurfaceError, TraceTIRError
from rayoptics.util.misc_math import normalize


RayPkg = namedtuple('RayPkg', ['ray', 'op', 'wvl'])
RaySeg = namedtuple('RaySeg', ['p', 'd', 'dst', 'nrml'])


def ray_pkg(ray_pkg):
    """ return a |Series| containing a ray package (RayPkg) """
    return pd.Series(ray_pkg, index=['ray', 'op', 'wvl'])


def ray_df(ray):
    """ return a |DataFrame| containing ray data """
    r = pd.DataFrame(ray, columns=['inc_pt', 'after_dir',
                                   'after_dst', 'normal'])
    r.index.names = ['intrfc']
    return r


@attr.s
class RSeg():
    inc_pt = attr.ib()
    after_dir = attr.ib()
    after_dst = attr.ib()
    normal = attr.ib()
    phase = attr.ib(default=0.0)


def list_ray(ray_obj, tfrms=None):
    """ pretty print a ray either in local or global coordinates """
    if type(ray_obj) is tuple:
        ray = ray_obj[0]
    else:
        ray = ray_obj

    colHeader = "            X            Y            Z           L" \
                "            M            N               Len"
    print(colHeader)

    colFormats = "{:3d}: {:12.5f} {:12.5f} {:12.5g} {:12.6f} {:12.6f} " \
                 "{:12.6f} {:12.5g}"

    for i, r in enumerate(ray):
        if tfrms is None:
            print(colFormats.format(i,
                                    r[mc.p][0], r[mc.p][1], r[mc.p][2],
                                    r[mc.d][0], r[mc.d][1], r[mc.d][2],
                                    r[mc.dst]))
        else:
            rot, trns = tfrms[i]
            p = rot.dot(r[mc.p]) + trns
            d = rot.dot(r[mc.d])
            print(colFormats.format(i, p[0], p[1], p[2], d[0], d[1], d[2],
                                    r[mc.dst]))


def trace(sequence, pt0, dir0, wvl, **kwargs):
    """ returns (ray, ray_opl, wvl)

    Args:
        sequence: a Sequence or generator that returns a list containing:
            Intfc, Gap, Trfm, Index, Z_Dir
        pt0: starting coordinate at object interface
        dir0: starting direction cosines following object interface
        wvl: ray trace wavelength in nm
        **kwargs: keyword arguments

    Returns:
        (**ray**, **op_delta**, **wvl**)

        - **ray** is a list for each interface in **path_pkg** of these
          elements: [pt, after_dir, after_dst, normal]

            - pt: the intersection point of the ray
            - after_dir: the ray direction cosine following the interface
            - after_dst: after_dst: the geometric distance to the next
              interface
            - normal: the surface normal at the intersection point

        - **op_delta** - optical path wrt equally inclined chords to the
          optical axis
        - **wvl** - wavelength (in nm) that the ray was traced in
    """
    return rt.trace(sequence, pt0, dir0, wvl, **kwargs)


def trace_base(opt_model, pupil, fld, wvl, **kwargs):
    """ returns (ray, ray_opl, wvl) """
    vig_pupil = fld.apply_vignetting(pupil)
    osp = opt_model.optical_spec
    fod = osp.parax_data.fod
    eprad = fod.enp_radius
    aim_pt = np.array([0., 0.])
    if hasattr(fld, 'aim_pt') and fld.aim_pt is not None:
        aim_pt = fld.aim_pt
    pt1 = np.array([eprad*vig_pupil[0]+aim_pt[0], eprad*vig_pupil[1]+aim_pt[1],
                    fod.obj_dist+fod.enp_dist])
    pt0 = osp.obj_coords(fld)
    dir0 = pt1 - pt0
    length = norm(dir0)
    dir0 = dir0/length
    return rt.trace(opt_model.seq_model, pt0, dir0, wvl, **kwargs)


def iterate_ray(opt_model, ifcx, xy_target, fld, wvl, **kwargs):
    """ iterates a ray to xy_target on interface ifcx, returns aim points on
    the paraxial entrance pupil plane

    If idcx is None, i.e. a floating stop surface, returns xy_target.

    If the iteration fails, a TraceError will be raised
    """
    def y_stop_coordinate(y1, *args):
        seq_model, ifcx, pt0, dist, wvl, y_target = args
        pt1 = np.array([0., y1, dist])
        dir0 = pt1 - pt0
        length = norm(dir0)
        dir0 = dir0/length
        try:
            ray, _, _ = rt.trace(seq_model, pt0, dir0, wvl)
        except TraceMissedSurfaceError as ray_miss:
            ray = ray_miss.ray
            if ray_miss.surf <= ifcx:
                raise ray_miss
        except TraceTIRError as ray_tir:
            ray = ray_tir.ray
            if ray_tir.surf < ifcx:
                raise ray_tir
        y_ray = ray[ifcx][mc.p][1]
#        print(y1, y_ray)
        return y_ray - y_target

    def surface_coordinate(coord, *args):
        seq_model, ifcx, pt0, dist, wvl, target = args
        pt1 = np.array([coord[0], coord[1], dist])
        dir0 = pt1 - pt0
        length = norm(dir0)
        dir0 = dir0/length
        ray, _, _ = rt.trace(seq_model, pt0, dir0, wvl)
        xy_ray = np.array([ray[ifcx][mc.p][0], ray[ifcx][mc.p][1]])
#        print(coord[0], coord[1], xy_ray[0], xy_ray[1])
        return xy_ray - target

    seq_model = opt_model.seq_model
    osp = opt_model.optical_spec

    fod = osp.parax_data.fod
    dist = fod.obj_dist + fod.enp_dist

    pt0 = osp.obj_coords(fld)
    if ifcx is not None:
        if pt0[0] == 0.0 and xy_target[0] == 0.0:
            # do 1D iteration if field and target points are zero in x
            y_target = xy_target[1]
            logging.captureWarnings(True)
            try:
                start_y, results = newton(y_stop_coordinate, 0.,
                                          args=(seq_model, ifcx, pt0,
                                                dist, wvl, y_target),
                                          disp=False, full_output=True)
            except RuntimeError as rte:
                # if we come here, start_y is a RuntimeResults object
                print(rte)
                start_y = results.root
            except TraceError:
                start_y = 0.0
            start_coords = np.array([0., start_y])
        else:
            # do 2D iteration. epsfcn is a parameter increment,
            #  make proportional to pupil radius
            try:
                start_coords = fsolve(surface_coordinate, np.array([0., 0.]),
                                      epsfcn=0.0001*fod.enp_radius,
                                      args=(seq_model, ifcx, pt0, dist,
                                            wvl, xy_target))
            except TraceError:
                start_coords = np.array([0., 0.])
    else:  # floating stop surface - use entrance pupil for aiming
        start_coords = np.array([0., 0.]) + xy_target

    return start_coords


def trace_with_opd(opt_model, pupil, fld, wvl, foc, **kwargs):
    """ returns (ray, ray_opl, wvl, opd) """
    ray_pkg = trace_base(opt_model, pupil, fld, wvl, **kwargs)

    rs_pkg, cr_pkg = setup_pupil_coords(opt_model, fld, wvl, foc)
    fld.chief_ray = cr_pkg
    fld.ref_sphere = rs_pkg

    opd, e1, ekp, ep = rt.wave_abr(fld, wvl, foc, ray_pkg)
    ray, ray_op, wvl = ray_pkg
    return ray, ray_op, wvl, opd


def trace_boundary_rays_at_field(opt_model, fld, wvl, use_named_tuples=False):
    """ returns a list of RayPkgs for the boundary rays for field fld
    """
    rim_rays = []
    osp = opt_model.optical_spec
    for p in osp.pupil.pupil_rays:
        try:
            ray, op, wvl = trace_base(opt_model, p, fld, wvl)
        except TraceError as ray_error:
            ray = ray_error.ray
            op = 0.

        if use_named_tuples:
            ray = [RaySeg(*rs) for rs in ray]
        rim_rays.append(RayPkg(ray, op, wvl))
    return rim_rays


def boundary_ray_dict(opt_model, rim_rays):
    pupil_rays = {}
    for ray, lbl in zip(rim_rays, opt_model.optical_spec.pupil.ray_labels):
        pupil_rays[lbl] = ray
    return pupil_rays


def trace_boundary_rays(opt_model, **kwargs):
    rayset = []
    wvl = opt_model.seq_model.central_wavelength()
    fov = opt_model.optical_spec.field_of_view
    for fi, fld in enumerate(fov.fields):
        rim_rays = trace_boundary_rays_at_field(opt_model, fld, wvl, **kwargs)
        fld.pupil_rays = boundary_ray_dict(opt_model, rim_rays)
        rayset.append(rim_rays)
    return rayset


def trace_ray_list_at_field(opt_model, ray_list, fld, wvl, foc):
    """ returns a list of ray |DataFrame| for the ray_list at field fld """
    rayset = []
    for p in ray_list:
        ray, op, wvl = trace_base(opt_model, p, fld, wvl)
        rayset.append(ray)
    rdf_list = [ray_df(r) for r in rayset]
    return rdf_list


def trace_field(opt_model, fld, wvl, foc):
    """ returns a |DataFrame| with the boundary rays for field fld """
    osp = opt_model.optical_spec
    pupil_rays = osp.pupil.pupil_rays
    rdf_list = trace_ray_list_at_field(opt_model, pupil_rays, fld, wvl, foc)
    rset = pd.concat(rdf_list, keys=osp.pupil.ray_labels,
                     names=['pupil'])
    return rset


def trace_all_fields(opt_model):
    """ returns a |DataFrame| with the boundary rays for all fields """
    osp = opt_model.optical_spec
    fld, wvl, foc = osp.lookup_fld_wvl_focus(0)
    fset = []
    for f in osp.field_of_view.fields:
        rset = trace_field(opt_model, f, wvl, foc)
        fset.append(rset)

    fdf = pd.concat(fset, keys=osp.field_of_view.index_labels,
                    names=['field'])
    return fdf


def trace_chief_ray(opt_model, fld, wvl, foc):
    osp = opt_model.optical_spec
    fod = osp.parax_data.fod

    ray, op, wvl = trace_base(opt_model, [0., 0.], fld, wvl)
    cr = RayPkg(ray, op, wvl)

    # cr_exp_pt: E upper bar prime: pupil center for pencils from Q
    # cr_exp_pt, cr_b4_dir, cr_exp_dist
    cr_exp_seg = rt.transfer_to_exit_pupil(opt_model.seq_model.ifcs[-2],
                                           (cr.ray[-2][mc.p],
                                            cr.ray[-2][mc.d]), fod.exp_dist)
    return cr, cr_exp_seg


def trace_fan(opt_model, fan_rng, fld, wvl, foc, img_filter=None,
              **kwargs):
    start = np.array(fan_rng[0])
    stop = fan_rng[1]
    num = fan_rng[2]
    step = (stop - start)/(num - 1)
    fan = []
    for r in range(num):
        pupil = np.array(start)
        ray_pkg = trace_base(opt_model, pupil, fld, wvl, **kwargs)

        if img_filter:
            result = img_filter(pupil, ray_pkg)
            fan.append([pupil, result])
        else:
            fan.append([pupil, ray_pkg])

        start += step
    return fan


def trace_grid(opt_model, grid_rng, fld, wvl, foc, img_filter=None,
               form='grid', append_if_none=True, **kwargs):
    start = np.array(grid_rng[0])
    stop = grid_rng[1]
    num = grid_rng[2]
    step = np.array((stop - start)/(num - 1))
    grid = []
    for i in range(num):
        if form == 'list':
            working_grid = grid
        elif form == 'grid':
            grid_row = []
            working_grid = grid_row

        for j in range(num):
            pupil = np.array(start)
            if (pupil[0]**2 + pupil[1]**2) < 1.0:
                ray_pkg = trace_base(opt_model, pupil, fld, wvl, **kwargs)
                if img_filter:
                    result = img_filter(pupil, ray_pkg)
                    working_grid.append(result)
                else:
                    working_grid.append([pupil[0], pupil[1], ray_pkg])
            else:  # ray outside pupil
                if img_filter:
                    result = img_filter(pupil, None)
                    if result is not None or append_if_none:
                        working_grid.append(result)
                else:
                    if append_if_none:
                        working_grid.append([pupil[0], pupil[1], None])

            start[1] += step[1]
        if form == 'grid':
            grid.append(grid_row)
        start[0] += step[0]
        start[1] = grid_rng[0][1]
    return np.array(grid)


def aim_chief_ray(opt_model, fld, wvl=None):
    """ aim chief ray at center of stop surface and save results on **fld** """
    seq_model = opt_model.seq_model
    if wvl is None:
        wvl = seq_model.central_wavelength()
    stop = seq_model.stop_surface
    aim_pt = iterate_ray(opt_model, stop, np.array([0., 0.]), fld, wvl)
    return aim_pt


def setup_pupil_coords(opt_model, fld, wvl, foc,
                       chief_ray_pkg=None, image_pt=None):
    if chief_ray_pkg is None:
        aim_chief_ray(opt_model, fld, wvl=wvl)
        chief_ray_pkg = trace_chief_ray(opt_model, fld, wvl, foc)
    elif chief_ray_pkg[2] != wvl:
        chief_ray_pkg = trace_chief_ray(opt_model, fld, wvl, foc)

    cr, cr_exp_seg = chief_ray_pkg

    if image_pt is None:
        image_pt = cr.ray[-1][mc.p]

    # cr_exp_pt: E upper bar prime: pupil center for pencils from Q
    # cr_exp_pt, cr_b4_dir, cr_dst
    cr_exp_pt = cr_exp_seg[mc.p]
    cr_exp_dist = cr_exp_seg[mc.dst]

    seq_model = opt_model.seq_model
    img_dist = seq_model.gaps[-1].thi
    img_pt = np.array(image_pt)
    img_pt[2] += img_dist

    # R' radius of reference sphere for O'
    ref_sphere_vec = img_pt - cr_exp_pt
    ref_sphere_radius = np.linalg.norm(ref_sphere_vec)
    ref_dir = normalize(ref_sphere_vec)

    ref_sphere = (image_pt, cr_exp_pt, cr_exp_dist,
                  ref_dir, ref_sphere_radius)

    z_dir = seq_model.z_dir[-1]
    wl = seq_model.index_for_wavelength(wvl)
    n_obj = seq_model.rndx[0][wl]
    n_img = seq_model.rndx[-1][wl]
    ref_sphere_pkg = (ref_sphere, opt_model.optical_spec.parax_data,
                      n_obj, n_img, z_dir)

    return ref_sphere_pkg, chief_ray_pkg


def setup_canonical_coords(opt_model, fld, wvl, image_pt=None):
    osp = opt_model.optical_spec
    seq_model = opt_model.seq_model
    fod = osp.parax_data.fod

    if fld.chief_ray is None:
        ray, op, wvl = trace_base(opt_model, [0., 0.], fld, wvl)
        fld.chief_ray = RayPkg(ray, op, wvl)
    cr = fld.chief_ray

    if image_pt is None:
        image_pt = cr.ray[-1][mc.p]

    # cr_exp_pt: E upper bar prime: pupil center for pencils from Q
    # cr_exp_pt, cr_b4_dir, cr_dst
    cr_exp_seg = rt.transfer_to_exit_pupil(seq_model.ifcs[-2],
                                           (cr.ray[-2][mc.p],
                                            cr.ray[-2][mc.d]),
                                           fod.exp_dist)
    cr_exp_pt = cr_exp_seg[mc.p]
    cr_exp_dist = cr_exp_seg[mc.dst]

    img_dist = seq_model.gaps[-1].thi
    img_pt = np.array(image_pt)
    img_pt[2] += img_dist

    # R' radius of reference sphere for O'
    ref_sphere_vec = img_pt - cr_exp_pt
    ref_sphere_radius = np.linalg.norm(ref_sphere_vec)
    ref_dir = normalize(ref_sphere_vec)

    ref_sphere = (image_pt, cr_exp_pt, cr_exp_dist,
                  ref_dir, ref_sphere_radius)

    z_dir = seq_model.z_dir[-1]
    wl = seq_model.index_for_wavelength(wvl)
    n_obj = seq_model.rndx[0][wl]
    n_img = seq_model.rndx[-1][wl]
    ref_sphere_pkg = (ref_sphere, osp.parax_data, n_obj, n_img, z_dir)
    fld.ref_sphere = ref_sphere_pkg
    return ref_sphere_pkg, cr


def trace_astigmatism_coddington_fan(opt_model, fld, wvl, foc):
    """ calculate astigmatism by Coddington trace at **fld** """
    cr = RayPkg(*trace_base(opt_model, [0., 0.], fld, wvl))
    s_dfoc, t_dfoc = trace_coddington_fan(opt_model, cr, foc=foc)
    return s_dfoc, t_dfoc


def trace_coddington_fan(opt_model, ray_pkg, foc=None):
    """ astigmatism calculation via Coddington trace

.. note:: spherical surfaces only
    """
    seq_model = opt_model.seq_model
    wl = seq_model.index_for_wavelength(ray_pkg.wvl)

    path = itertools.zip_longest(ray_pkg.ray, seq_model.ifcs,
                                 [n[wl] for n in seq_model.rndx],
                                 seq_model.lcl_tfrms,
                                 seq_model.z_dir)

    before_rind = seq_model.rndx[0][wl]
    before_dir = None
    for r, ifc, after_rind, tfrm, z_dir in path:
        pt, after_dir, after_dst, normal = r
        if before_dir is not None:
            normal_len = norm(normal)
            cosI_prime = np.dot(after_dir, normal)/normal_len
            sinI_prime = math.sqrt(1.0 - cosI_prime**2)
            sinI = after_rind*sinI_prime/before_rind
            cosI = math.sqrt(1.0 - sinI**2)

            obl_power = ifc.optical_power
            if obl_power != 0.0:
                obl_power *= ((after_rind*cosI_prime - before_rind*cosI) /
                              (after_rind - before_rind))
#                print("pwr, obl_pwr, after_dst:",
#                      ifc.optical_power/(after_rind - before_rind),
#                      obl_power, after_dst)
#            else:
#                print("pwr, obl_pwr, after_dst:",
#                      ifc.optical_power, obl_power, after_dst)

            n_by_s_prime = before_rind/s_before + obl_power
            s_prime = after_rind/n_by_s_prime
#            print("s, s':", s_before, s_prime)
            s_before = s_prime - after_dst

            n_cosIp2_by_t_prime = before_rind*cosI**2/t_before + obl_power
            t_prime = after_rind*cosI_prime**2/n_cosIp2_by_t_prime
#            print("t, t':", t_before, t_prime)
            t_before = t_prime - after_dst
        else:
            s_before = -after_dst
            t_before = -after_dst

        before_rind = after_rind
        before_dir = after_dir

    s_dfoc = s_prime*after_dir[2] + pt[2]
    t_dfoc = t_prime*after_dir[2] + pt[2]
    if foc is not None:
        focus_shift = foc
        s_dfoc -= focus_shift
        t_dfoc -= focus_shift

#    print("delta s, t:", s_dfoc, t_dfoc)
    return s_dfoc, t_dfoc


def intersect_2_lines(P1, V1, P2, V2):
    """ intersect 2 non-parallel lines, returning distance from P1

    s = ((P2 - P1) x V1).(V1 x V2)/\|(V1 x V2)\|**2

    `Weisstein, Eric W. "Line-Line Intersection." From MathWorld--A Wolfram Web
    Resource. <http://mathworld.wolfram.com/Line-LineIntersection.html>`_
    """
    Vx = np.cross(V1, V2)
    s = np.dot(np.cross(P2 - P1, V1), Vx)/np.dot(Vx, Vx)
    return s


def trace_astigmatism(opt_model, fld, wvl, foc, dx=0.001, dy=0.001):
    """ calculate astigmatism by tracing close rays about the chief ray at **fld**

    This function implicitly assumes that the **fld** point is on a plane of
    symmetry, i.e. the system is rotationally symmetric, bilaterally symmetric,
    or quad symmetric. No check is done to ensure this.

    Args:
        opt_model: the optical model
        fld: a Field object
        wvl: wavelength in nm
        foc: defocus amount
        dx: delta in pupil coordinates for x/sagittal direction
        dy: delta in pupil coordinates for y/tangential direction

    Returns:
        tuple: sagittal and tangential focus shifts at **fld**
    """
    rlist = []
    rlist.append(RayPkg(*trace_base(opt_model, [0., 0.], fld, wvl)))
    rlist.append(RayPkg(*trace_base(opt_model, [dx, 0.], fld, wvl)))
    rlist.append(RayPkg(*trace_base(opt_model, [0., dy], fld, wvl)))
    rlist.append(RayPkg(*trace_base(opt_model, [-dx, 0.], fld, wvl)))
    rlist.append(RayPkg(*trace_base(opt_model, [0., -dy], fld, wvl)))

    s = intersect_2_lines(rlist[1].ray[-1][mc.p], rlist[1].ray[-1][mc.d],
                          rlist[3].ray[-1][mc.p], rlist[3].ray[-1][mc.d])
    s_foc = s * rlist[1].ray[-1][mc.d][2]

    t = intersect_2_lines(rlist[2].ray[-1][mc.p], rlist[2].ray[-1][mc.d],
                          rlist[4].ray[-1][mc.p], rlist[4].ray[-1][mc.d])
    t_foc = t * rlist[2].ray[-1][mc.d][2]

    if foc is not None:
        focus_shift = foc
        s_foc -= focus_shift
        t_foc -= focus_shift
    return s_foc, t_foc
